/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort;

import dk.sdsd.nsp.slalog.api.SLALogConfig;
import dk.sdsd.nsp.slalog.api.SLALogger;
import dk.stamkort.health.LastStateHealthIndicator;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.integrations.*;
import dk.stamkort.integrations.dds.*;
import dk.stamkort.integrations.minlog.MinLogClient;
import dk.stamkort.integrations.minlog.MinLogException;
import dk.stamkort.util.Properties;
import dk.stamkort.util.SecretProperties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.actuate.health.HealthIndicator;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.context.annotation.Bean;
import org.springframework.ws.config.annotation.WsConfigurerAdapter;
import org.springframework.ws.soap.SoapVersion;
import org.springframework.ws.soap.saaj.SaajSoapMessageFactory;

import javax.annotation.PostConstruct;
import javax.xml.bind.JAXBException;
import java.net.MalformedURLException;
import java.time.Instant;

@SpringBootApplication
public class PersonalDataCardApplication extends WsConfigurerAdapter {

    private static final Logger log = LogManager.getLogger(PersonalDataCardApplication.class);

    private static final String IT_SYSTEM_NAME = "Fælles Stamkort";

    public static Instant DEPLOY_TIME = Instant.now();
    public static final String MINLOG_CONFIG_FILE_NAME = "minlogclient.properties";

    public static void main(String[] args) {
        SpringApplication.run(PersonalDataCardApplication.class, args);
    }

    @PostConstruct
    public void logSettings() {
        log.info("Info logging is enabled.");
        log.debug("Debug logging is enabled.");
        log.trace("Trace logging is enabled.");
    }

    @Bean
    public MinLogClient fskMinLogClient() throws MinLogException {
        return new MinLogClient("/" + MINLOG_CONFIG_FILE_NAME);
    }

    @Bean
    public OrganDonorClient organDonorClient(Properties properties) throws JAXBException {
        return new OrganDonorClient(
                properties.getStsEndpoint(),
                properties.getOdrEndpoint(),
                properties.getOdrConnectTimeoutMillis(),
                properties.getOdrReadTimeoutMillis());
    }

    @Bean
    public LivingWillClient livingWillClient(Properties properties) throws JAXBException {
        return new LivingWillClient(
                properties.getStsEndpoint(),
                properties.getLtrEndpoint(),
                properties.getLtrConnectTimeoutMillis(),
                properties.getLtrReadTimeoutMillis());
    }

    @Bean
    public TreatmentWillClient treatmentWillClient(Properties properties) throws JAXBException {
        return new TreatmentWillClient(
                properties.getStsEndpoint(),
                properties.getBtrEndpoint(),
                properties.getBtrConnectTimeoutMillis(),
                properties.getBtrReadTimeoutMillis());
    }

    @Bean
    public PersonalDataCardRegisterClient personalDataCardRegisterClient(Properties properties) throws JAXBException {
        return new PersonalDataCardRegisterClient(
                properties.getStsEndpoint(),
                properties.getSkrEndpoint(),
                properties.getSkrConnectTimeoutMillis(),
                properties.getSkrReadTimeoutMillis());
    }

    @Bean
    public YderClient yderClient(Properties properties) throws JAXBException {
        return new YderClient(
                properties.getStsEndpoint(),
                properties.getSyesEndpoint(),
                properties.getSyesConnectTimeoutMillis(),
                properties.getSyesReadTimeoutMillis());
    }


    @Bean
    public SCESClient scesClient(Properties properties, SecretProperties secretProperties) {
        return new SCESClient(
                properties.getClientKeystorePath(),
                secretProperties.getClientKeystorePassword(),
                properties.getClientOrgId(),
                properties.getClientOrgName(),
                properties.getStsEndpoint(),
                properties.getScesEndpoint(),
                properties.getScesConnectTimeoutMillis(),
                properties.getScesReadTimeoutMillis(),
                IT_SYSTEM_NAME,
                properties.isProduction());
    }

    @Bean
    public DDSRegisterOnDemandClient ddsOnDemandClient(Properties properties, SecretProperties secretProperties) throws MalformedURLException {
        return new DDSRegisterOnDemandClientImpl(
                properties.getClientKeystorePath(),
                secretProperties.getClientKeystorePassword(),
                properties.getClientOrgId(),
                properties.getClientOrgName(),
                properties.getStsEndpoint(),
                properties.getDdsRegisterOnDemandURL(),
                properties.getDdsConnectTimeoutMillis(),
                properties.getDdsReadTimeoutMillis(),
                IT_SYSTEM_NAME,
                properties.isProduction());
    }

    @Bean
    public DDSRegistryStoredQueryClient ddsRegistryStoredQueryClient(Properties properties, SecretProperties secretProperties) throws MalformedURLException {
        return new DDSRegistryStoredQueryClientImpl(
                properties.getClientKeystorePath(),
                secretProperties.getClientKeystorePassword(),
                properties.getClientOrgId(),
                properties.getClientOrgName(),
                properties.getStsEndpoint(),
                properties.getDdsRegistryStoredQueryURL(),
                properties.getDdsConnectTimeoutMillis(),
                properties.getDdsReadTimeoutMillis(),
                IT_SYSTEM_NAME,
                properties.isProduction());
    }

    @Bean
    public DDSRegistryUpdateClient ddsRegistryUpdateClient(Properties properties, SecretProperties secretProperties) throws MalformedURLException {
        return new DDSRegistryUpdateClientImpl(
                properties.getClientKeystorePath(),
                secretProperties.getClientKeystorePassword(),
                properties.getClientOrgId(),
                properties.getClientOrgName(),
                properties.getStsEndpoint(),
                properties.getDdsRegistryUpdateURL(),
                properties.getDdsConnectTimeoutMillis(),
                properties.getDdsReadTimeoutMillis(),
                IT_SYSTEM_NAME,
                properties.isProduction());
    }

    @Bean
    public SLALogger slaLogger() {
        return new SLALogConfig("Fælles stamkort", "fsk").getSLALogger();
    }

    @Bean
    public SaajSoapMessageFactory messageFactory() {
        SaajSoapMessageFactory messageFactory = new SaajSoapMessageFactory();
        messageFactory.setSoapVersion(SoapVersion.SOAP_12);
        return messageFactory;
    }

    /*
     * Actuator HealthIndicator beans:
     * This HealthIndicator beans are automatically added to Actuator and the name of the bean (except for the HealthIndicator
     * suffix) is used as the entry name in the health information.
     */

    @Bean
    public HealthIndicator organDonorClientHealthIndicator(@Qualifier("organDonorClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator livingWillClientHealthIndicator(@Qualifier("livingWillClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator treatmentWillClientHealthIndicator(@Qualifier("treatmentWillClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator personalDataCardRegisterClientHealthIndicator(@Qualifier("personalDataCardRegisterClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator yderClientHealthIndicator(@Qualifier("yderClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }


    @Bean
    public HealthIndicator scesClientHealthIndicator(@Qualifier("scesClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator minLogClientHealthIndicator(@Qualifier("fskMinLogClient") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }

    @Bean
    public HealthIndicator syncJobHealthIndicator(@Qualifier("syncJob") LastStateHealthReporter reporter) {
        return LastStateHealthIndicator.reporter(reporter).build();
    }
}
