/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.xml.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.xml.bind.*;
import javax.xml.namespace.QName;
import javax.xml.soap.*;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

public class DGWSClient {
    protected String endpoint;
    protected TransportUtil transport;

    private SOSIFactory sosiFactory;
    private SystemIDCard idCard;
    private static Map<Class<?>, JAXBContext> jaxbContextMap = new HashMap<>();

    public DGWSClient(String keystorePath, String keystorePassword, String orgId, String orgName, String stsEndpoint, String endpoint, int connectTimeoutMillis, int readTimeoutMillis, String itSystemName, boolean isProduction) {
        this.transport = new TransportUtil(keystorePath, keystorePassword, orgId, orgName, stsEndpoint, connectTimeoutMillis, readTimeoutMillis, itSystemName, isProduction);
        this.endpoint = endpoint;
    }

    protected <T> Element toXml(T body, QName bodyQname) throws JAXBException {
        JAXBElement<T> bodyJaxbElm = new JAXBElement<>(bodyQname, (Class<T>) body.getClass(), body);
        Document requestDom = XmlUtil.createEmptyDocument();

        getMarshaller(body.getClass()).marshal(bodyJaxbElm, requestDom);
        return requestDom.getDocumentElement();
    }

    protected SOAPBody getSoapBody(String xmlInput) throws SOAPException, IOException {
        MessageFactory factory = MessageFactory.newInstance();
        SOAPMessage message = factory.createMessage(
                new MimeHeaders(),
                new ByteArrayInputStream(xmlInput.getBytes(StandardCharsets.UTF_8)));
        return message.getSOAPBody();
    }

    protected Request prepareRequest() {
        if (idCard == null || !idCard.isValidInTime()) {
            if (sosiFactory == null) {
                sosiFactory = transport.createSOSIFactory(false); // validate=false because DDS error are returned without a SOAP header
            }
            idCard = transport.getStsSignedIdCard(sosiFactory);
        }
        Request request = sosiFactory.createNewRequest(false, null);
        request.setIDCard(idCard);
        return request;
    }

    public Marshaller getMarshaller(Class<?> c) throws JAXBException {
        return getJAXBContext(c).createMarshaller();
    }

    public Unmarshaller getUnmarshaller(Class<?> c) throws JAXBException {
        return getJAXBContext(c).createUnmarshaller();
    }

    private JAXBContext getJAXBContext(Class<?> c) throws JAXBException {
        JAXBContext context = jaxbContextMap.get(c);
        if (context == null) {
            context = JAXBContext.newInstance(c);
            jaxbContextMap.put(c, context);
        }
        return context;
    }
}
