/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.medcom.dgws._2006._04.dgws_1_0.Linking;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.SlaLogger;
import org.apache.commons.io.Charsets;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Element;

import javax.xml.bind.*;
import javax.xml.namespace.QName;
import javax.xml.soap.*;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.StringWriter;
import java.time.Instant;
import java.util.UUID;

public abstract class ReusingDGWSClient<REQUEST, RESPONSE> implements LastStateHealthReporter {
    private static final Logger log = LogManager.getLogger(ReusingDGWSClient.class);

    private String name;
    private Class<REQUEST> requestClass;
    private Class<RESPONSE> responseClass;
    private SlaLogger slaLogger;
    private String endpoint;
    private TransportUtil transport;

    private Unmarshaller soapBodyUnmarshaller;

    private Unmarshaller medcomHeaderUnmarshaller;
    private Marshaller medcomHeaderMarshaller;

    private Boolean lastExecutionSucceeded;
    private Instant timeOfLastExecution;
    private Exception lastException;
    private Marshaller marshaller;

    public ReusingDGWSClient(String name, Class<REQUEST> requestClass, Class<RESPONSE> responseClass, String stsEndpoint, String endpoint, int connectTimeoutMillis, int readTimeoutMillis) throws JAXBException {
        this.name = name;
        this.requestClass = requestClass;
        this.responseClass = responseClass;
        this.slaLogger = new SlaLogger(name);
        this.transport = new TransportUtil(stsEndpoint, connectTimeoutMillis, readTimeoutMillis);
        this.endpoint = endpoint;

        JAXBContext medcomHeaderJaxbContext = JAXBContext.newInstance(Header.class);
        medcomHeaderUnmarshaller = medcomHeaderJaxbContext.createUnmarshaller();
        medcomHeaderMarshaller = medcomHeaderJaxbContext.createMarshaller();

        JAXBContext jaxbContextRequest = JAXBContext.newInstance(requestClass);
        marshaller = jaxbContextRequest.createMarshaller();

        JAXBContext jaxbContextResponse = JAXBContext.newInstance(responseClass);
        soapBodyUnmarshaller = jaxbContextResponse.createUnmarshaller();
    }

    public RESPONSE call(REQUEST request, RequestContextVO requestContext) throws DGWSClientException {

        JAXBElement<REQUEST> bodyJaxbElm = new JAXBElement<>(getRequestQName(), requestClass, request);

        return call(bodyJaxbElm, requestContext);
    }

    public RESPONSE call(JAXBElement<REQUEST> request, RequestContextVO requestContext) throws DGWSClientException {
        long startMillis = System.currentTimeMillis();
        long startNanos = System.nanoTime();
        try {
            String responseXml = callInternal(request, requestContext);
            addSlaLog(requestContext, startNanos, startMillis, null);

            RESPONSE response = fromXml(responseXml);

            lastExecutionSucceeded = true;
            timeOfLastExecution = Instant.now();

            return response;
        } catch (Exception e) {
            String errorMessage = "Error calling " + name;
            String msg = e.getMessage();
            if (msg != null) {
                errorMessage += ":" + msg;
            }
            addSlaLog(requestContext, startNanos, startMillis, errorMessage);

            lastExecutionSucceeded = false;
            timeOfLastExecution = Instant.now();
            lastException = e;

            throw new DGWSClientException(errorMessage, e);
        }
    }

    public String callInternal(JAXBElement<REQUEST> request, RequestContextVO requestContext) throws IOException, JAXBException, SOAPException, TransformerException {

        // create SOAPMessage with request body
        MessageFactory messageFactory = MessageFactory.newInstance();
        SOAPMessage soapMessage = messageFactory.createMessage();
        SOAPEnvelope soapEnvelope = soapMessage.getSOAPPart().getEnvelope();
        SOAPBody soapBody = soapEnvelope.getBody();

        marshaller.marshal(request, soapBody);

        SOAPHeader soapHeader = soapEnvelope.getHeader();
        Element medcomHeaderElement = requestContext.getMedcomHeader();

        // insert security header
        if (requestContext.getSecurityHeader() != null) {
            soapHeader.appendChild(soapHeader.getOwnerDocument().importNode(requestContext.getSecurityHeader(), true));
        }

        // insert medcom header with adjusted MessageId's and FlowId's
        if (medcomHeaderElement != null) {
            String messageId = UUID.randomUUID().toString(); // separate messageId for each DGWS call
            String flowId = requestContext != null && requestContext.getProxyData() != null ? requestContext.getProxyData().getFlowId() : null; // same flowId for each DGWS call

            Header medcomHeader = (Header) medcomHeaderUnmarshaller.unmarshal(medcomHeaderElement);
            Linking linking = medcomHeader.getLinking();
            if (linking == null) {
                linking = new Linking();
                medcomHeader.setLinking(linking);
            }

            linking.setMessageID(messageId);
            linking.setFlowID(flowId);

            medcomHeaderMarshaller.marshal(medcomHeader, soapHeader);
        }

        // get xml as string with correct encoding
        StringWriter sw = new StringWriter();
        TransformerFactory.newInstance().newTransformer().transform(new DOMSource(soapMessage.getSOAPPart()), new StreamResult(sw));
        String requestXml = sw.toString();

        log.debug(name + " request: " + requestXml);
        String responseXml = transport.sendRequest(endpoint, getSOAPAction(), requestXml);
        log.debug(name + " response: " + responseXml);

        return responseXml;
    }

    protected abstract QName getRequestQName();

    protected abstract String getSOAPAction();

    private RESPONSE fromXml(String xml) throws SOAPException, IOException, JAXBException, DGWSClientException {
        SOAPMessage soapMessage = MessageFactory.newInstance().createMessage(null, new ByteArrayInputStream(xml.getBytes(Charsets.UTF_8)));
        SOAPBody soapBody = soapMessage.getSOAPBody(); // throws an exception if body is malformed
        log.debug("SOAPMessage parsed successfully");

        if (soapBody.hasFault()) {
            SOAPFault fault = soapBody.getFault();
            throw new DGWSClientException("Service " + name + " returned fault: " + fault.getFaultString());
        }
        JAXBElement<RESPONSE> jaxbResponse = soapBodyUnmarshaller.unmarshal(soapBody.getFirstChild(), responseClass);

        return jaxbResponse.getValue();
    }

    private void addSlaLog(RequestContextVO requestContext, long startNanos, long startMillis, String errorMessage) {
        if (slaLogger != null) {
            slaLogger.addSlaLog(requestContext, startNanos, startMillis, errorMessage);
        }
    }

    @Override
    public Boolean lastExecutionSucceeded() {
        return lastExecutionSucceeded;
    }

    @Override
    public Instant timeOfLastExecution() {
        return timeOfLastExecution;
    }

    @Override
    public Exception lastException() {
        return lastException;
    }
}
