/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.ModelException;
import dk.sosi.seal.model.Reply;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.xml.XmlUtil;
import dk.sosi.seal.xml.XmlUtilException;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.SlaLogger;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationIn;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.soap.*;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.time.Instant;

public class SCESClient extends DGWSClient implements LastStateHealthReporter {
    private static final Logger log = LogManager.getLogger(SCESClient.class);

    private oio.medcom.cprservice._1_0.ObjectFactory objectFactory = new oio.medcom.cprservice._1_0.ObjectFactory();
    private SlaLogger slaLogger = new SlaLogger("SCES");
    private SOSIFactory sosiFactory;

    private Boolean lastExecutionSucceeded;
    private Instant timeOfLastExecution;
    private Exception lastException;

    public SCESClient(String keystorePath, String keystorePassword, String orgId, String orgName, String stsEndpoint, String cpreEndpoint, int connectTimeoutMillis, int readTimeoutMillis, String itSystemName, boolean isProduction) {
        super(keystorePath, keystorePassword, orgId, orgName, stsEndpoint, cpreEndpoint, connectTimeoutMillis, readTimeoutMillis, itSystemName, isProduction);
        this.endpoint = cpreEndpoint;
    }

    public SCESClient setTransportUtil(TransportUtil transport) {
        super.transport = transport;
        return this;
    }

    public GetPersonWithHealthCareInformationOut call(String cpr, RequestContextVO context) throws DGWSClientException {
        long startMillis = System.currentTimeMillis();
        long startNanos = System.nanoTime();
        try {
            // Call CPRE service. Throws FSKException if no data exists for the CPR in question
            if(sosiFactory == null) {
                sosiFactory = transport.createSOSIFactory(true);
            }
            String flowId = context != null && context.getProxyData() != null ? context.getProxyData().getFlowId() : null;
            String responseXml = callInternal(sosiFactory, cpr, flowId);
            addSlaLog(context, startNanos, startMillis, null);
            GetPersonWithHealthCareInformationOut person = fromXml(sosiFactory, responseXml, cpr);

            lastExecutionSucceeded = true;
            timeOfLastExecution = Instant.now();

            return person;
        } catch (IOException | JAXBException | SOAPException | XmlUtilException | ModelException | CredentialVaultException e) {
            String errorMessage = "Error calling CPRE";
            String msg = e.getMessage();
            if (msg != null) {
                errorMessage += ":" + msg;
            }
            addSlaLog(context, startNanos, startMillis, errorMessage);

            lastExecutionSucceeded = false;
            timeOfLastExecution = Instant.now();
            lastException = e;

            throw new DGWSClientException(errorMessage, e);
        }
    }

    public String callInternal(SOSIFactory sosiFactory, String cpr, String flowId) throws IOException, JAXBException, SOAPException {
        SystemIDCard idCard = transport.getStsSignedIdCard(sosiFactory);

        Request request = sosiFactory.createNewRequest(false, flowId);
        request.setIDCard(idCard);

        GetPersonWithHealthCareInformationIn getPersonWithHealthCareInformationIn = objectFactory.createGetPersonWithHealthCareInformationIn();
        getPersonWithHealthCareInformationIn.setPersonCivilRegistrationIdentifier(cpr);
        request.setBody(toXml(getPersonWithHealthCareInformationIn, new QName("urn:oio:medcom:cprservice:1.0.4", "getPersonWithHealthCareInformationIn")));

        Document soapRequestDOM = request.serialize2DOMDocument();
        String requestXml = XmlUtil.node2String(soapRequestDOM);

        log.debug("CPRE request: " + requestXml);
        String responseXml = transport.sendRequest(endpoint, "http://rep.oio.dk/medcom.sundcom.dk/xml/wsdl/2007/06/28/getPersonWithHealthCareInformation", requestXml);
        log.debug("CPRE response: " + responseXml);

        return responseXml;
    }

    private GetPersonWithHealthCareInformationOut fromXml(SOSIFactory sosiFactory, String responseXml, String cpr) throws JAXBException, SOAPException, IOException {
        MessageFactory factory = MessageFactory.newInstance();
        SOAPMessage message = factory.createMessage(new MimeHeaders(), new ByteArrayInputStream(responseXml.getBytes(Charset.forName("UTF-8"))));
        SOAPFault fault = message.getSOAPBody().getFault();
        String NO_DATA = "Ingen data fundet";
        if (fault != null) {
            if (NO_DATA.equals(fault.getFaultString())) {
                // We don't return PersonalDataCard for people not in CPR
                log.warn("No SCES data found for person: " + cpr);
                FaultUtil.throwFSKFault(FaultCode.NO_SCES_DATA_FOUND);
            } else {
                // Throw exception for CPRE faults
                FaultUtil.throwFSKFault(FaultCode.CALL_SCES_ERROR, fault.getFaultString());
            }
        }
        Reply reply = sosiFactory.deserializeReply(responseXml);
        Element body = reply.getBody();
        JAXBElement<GetPersonWithHealthCareInformationOut> jaxbResponse = getUnmarshaller(GetPersonWithHealthCareInformationOut.class).unmarshal(body, GetPersonWithHealthCareInformationOut.class);
        return jaxbResponse.getValue();
    }

    private void addSlaLog(RequestContextVO context, long startNanos, long startMillis, String errorMessage) {
        if (slaLogger != null) {
            slaLogger.addSlaLog(context, startNanos, startMillis, errorMessage);
        }
    }

    @Override
    public Boolean lastExecutionSucceeded() {
        return lastExecutionSucceeded;
    }

    @Override
    public Instant timeOfLastExecution() {
        return timeOfLastExecution;
    }

    @Override
    public Exception lastException() {
        return lastException;
    }
}
