/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.vault.CredentialVaultException;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.commons.io.IOUtils;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicHeader;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.security.cert.X509Certificate;
import java.util.Properties;

public class TransportUtil {
    private static final Logger log = LogManager.getLogger(TransportUtil.class);

    private static final Charset UTF_8 = Charset.forName("utf-8");
    private static final String orgType = SubjectIdentifierTypeValues.CVR_NUMBER;

    private static final Properties props = createSOSIProperties();

    private static Properties createSOSIProperties() {
        System.setProperty(SOSIFactory.PROPERTYNAME_SOSI_DO_NOT_REGISTER_STR_TRANSFORM, Boolean.TRUE.toString());
        Properties props = SignatureUtil.setupCryptoProviderForJVM();
        return props;
    }

    private CredentialVault vault = null;

    private String itSystemName;
    private String keystorePath;
    private String keystorePassword;
    private String orgId;
    private String orgName;
    private String stsEndpoint;
    private int connectTimeoutMillis;
    private int readTimeoutMillis;
    private boolean isProduction;

    public TransportUtil(String keystorePath, String keystorePassword, String orgId, String orgName, String stsEndpoint, int connectTimeoutMillis, int readTimeoutMillis, String itSystemName, boolean isProduction) {
        this.keystorePath = keystorePath;
        this.keystorePassword = keystorePassword;
        this.orgId = orgId;
        this.orgName = orgName;
        this.stsEndpoint = stsEndpoint;
        this.itSystemName = itSystemName;
        this.connectTimeoutMillis = connectTimeoutMillis;
        this.readTimeoutMillis = readTimeoutMillis;
        this.isProduction = isProduction;
    }

    public TransportUtil(String stsEndpoint, int connectTimeoutMillis, int readTimeoutMillis) {
        this.stsEndpoint = stsEndpoint;
        this.connectTimeoutMillis = connectTimeoutMillis;
        this.readTimeoutMillis = readTimeoutMillis;
    }

    public SOSIFactory createSOSIFactory(boolean validate) {
        props.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(validate));

        if (vault == null) {
            try {
                vault = loadKeyStore(props);
                if (vault == null) {
                    throw new CredentialVaultException("Keystore could not be loaded: " + keystorePath);
                }
            } catch (CredentialVaultException e) {
                log.error("Keystore could not be loaded. File path: " + keystorePath);
                throw e;
            }
        }

        Federation federation = isProduction ? new SOSIFederation(props) : new SOSITestFederation(props);
        SOSIFactory sosiFactory = new SOSIFactory(federation, vault, props);

        return sosiFactory;
    }

    private CredentialVault loadKeyStore(Properties props) throws CredentialVaultException {
        CredentialVault vault = new ClasspathCredentialVault(props, keystorePath, keystorePassword);
        log.info("Loaded keystore '" + keystorePath + "' from classpath");
        return vault;
    }

    /**
     * Constructs a FOCES signed ID-card and requests the STS to sign it. For the purpose of calling SAES, one could
     * have used a VOCES to initially sign.
     */
    public SystemIDCard getStsSignedIdCard(SOSIFactory sosiFactory) {
        // Null since we do not create level 2 id cards in this example
        String idcard_username = null;
        // Null since we do not create level 2 id cards in this example
        String idcard_password = null;
        // Null since we let Seal deside what our nameid should be
        String idcard_alternative_identifier = null;

        CareProvider careProvider = new CareProvider(orgType, orgId, orgName);
        X509Certificate certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();
        SystemIDCard systemIDCard = sosiFactory.createNewSystemIDCard(itSystemName, careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, idcard_username, idcard_password, certificate, idcard_alternative_identifier);
        return getSTSSignedIdCard(sosiFactory, systemIDCard, stsEndpoint);
    }

    private SystemIDCard getSTSSignedIdCard(SOSIFactory sosiFactory, SystemIDCard providedIdCard, String sts_endpoint) {
        SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
        securityTokenRequest.setIDCard(providedIdCard);
        Document doc = securityTokenRequest.serialize2DOMDocument();
        String requestXml = XmlUtil.node2String(doc, false, true);
        String responseXml;

        try {
            responseXml = sendRequest(sts_endpoint, "Issue", requestXml);
        } catch (IOException e) {
            log.warn(e);
            return null;
        }

        SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);
        SystemIDCard stsSignedIdCard = (SystemIDCard) securityTokenResponse.getIDCard();
        if (!securityTokenResponse.isFault() && stsSignedIdCard != null) {
            return stsSignedIdCard;
        } else {
            log.error("We got a SOAP Fault from STS. The detailed message can be seen in the xml response");
            if (securityTokenResponse.isFault()) {
                log.error("A SOAP Fault was received: " + securityTokenResponse.getFaultString());
            }

            return null;
        }
    }

    public String sendRequest(String url, String soapAction, String postBody) throws IOException {
        DefaultHttpClient defaultHttpClient = new DefaultHttpClient();
        defaultHttpClient.getParams().setParameter("http.connection.timeout", connectTimeoutMillis);
        defaultHttpClient.getParams().setParameter("http.socket.timeout", readTimeoutMillis);
        HttpPost httpPost = new HttpPost(url);
        httpPost.addHeader(new BasicHeader("Content-Type", "text/xml; charset=utf-8"));
        httpPost.addHeader(new BasicHeader("SOAPAction", "\"" + soapAction + "\""));
        //httpPost.addHeader(new BasicHeader("X-NSI-GW-PARTITION", "test"));
        //httpPost.addHeader(new BasicHeader("SOSIGW_PARTITION", "test"));
        httpPost.setEntity(new StringEntity(postBody, UTF_8));
        HttpResponse response = defaultHttpClient.execute(httpPost);
        String result = "";
        int statusCode = response.getStatusLine().getStatusCode();
        if (statusCode != 200 && statusCode != 500) {
            if (statusCode == 202) {
                return "";
            } else {
                result = response.getStatusLine().getReasonPhrase();
                throw new IOException("HTTP POST failed (" + statusCode + "): " + result);
            }
        } else {
            InputStream is = null;
            try {
                is = response.getEntity().getContent();
                result = IOUtils.toString(is, UTF_8);
            } finally {
                if (is != null) {
                    is.close();
                }
            }

            return result;
        }
    }
}
