/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.dds;

import dk.sosi.seal.model.Request;
import dk.sosi.seal.xml.XmlUtil;
import dk.stamkort.dao.exceptions.DDSPermanentException;
import dk.stamkort.dao.exceptions.DDSTemporaryException;
import dk.stamkort.integrations.DDSClient;
import dk.stamkort.util.Properties;
import oasis.names.tc.ebxml_regrep.xsd.query._3.AdhocQueryRequest;
import oasis.names.tc.ebxml_regrep.xsd.query._3.AdhocQueryResponse;
import oasis.names.tc.ebxml_regrep.xsd.query._3.ResponseOptionType;
import oasis.names.tc.ebxml_regrep.xsd.rim._3.AdhocQueryType;
import oasis.names.tc.ebxml_regrep.xsd.rim._3.IdentifiableType;
import oasis.names.tc.ebxml_regrep.xsd.rim._3.SlotType1;
import oasis.names.tc.ebxml_regrep.xsd.rim._3.ValueListType;
import org.springframework.beans.factory.annotation.Autowired;
import org.w3c.dom.Document;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.List;

public class DDSRegistryStoredQueryClientImpl extends DDSClient implements DDSRegistryStoredQueryClient {

    @Autowired
    Properties properties;

    private String ADHOC_QUERY_ID = "urn:uuid:5c4f972b-d56b-40ac-a5fc-c8ca9b40b9d4";

    private static final String STATUS_SUCCESS = "urn:oasis:names:tc:ebxml-regrep:ResponseStatusType:Success";

    public DDSRegistryStoredQueryClientImpl(String keystorePath, String keystorePassword, String orgId, String orgName, String stsEndpoint, String endpoint, int connectTimeoutMillis, int readTimeoutMillis, String itSystemName, boolean isProduction) throws MalformedURLException {
        super(DDSRegistryStoredQueryClientImpl.class.getName(), keystorePath, keystorePassword, orgId, orgName, stsEndpoint, endpoint, connectTimeoutMillis, readTimeoutMillis, itSystemName, isProduction);
    }

    public AdhocQueryResponse getDocumentById(String documentId) throws DDSTemporaryException {
        try {
            AdhocQueryRequest query = new AdhocQueryRequest();
            ResponseOptionType responseOption = new ResponseOptionType();
            responseOption.setReturnComposedObjects(true);
            responseOption.setReturnType("LeafClass");
            query.setResponseOption(responseOption);
            AdhocQueryType adhocQueryType = new AdhocQueryType();
            adhocQueryType.setId(ADHOC_QUERY_ID);
            SlotType1 uniqueIdSlot = new SlotType1();
            uniqueIdSlot.setName("$XDSDocumentEntryUniqueId");
            ValueListType valueList = new ValueListType();
            valueList.getValue().add("('" + documentId +"')");
            uniqueIdSlot.setValueList(valueList);
            adhocQueryType.getSlot().add(uniqueIdSlot);
            query.setAdhocQuery(adhocQueryType);

            Request request = prepareRequest();
            request.setBody(toXml(query, new QName("urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0", "AdhocQueryRequest")));
            Document soapRequestDOM = request.serialize2DOMDocument();
            String requestXml = XmlUtil.node2String(soapRequestDOM);
            String responseXml = sendRequest(requestXml, "urn:ihe:iti:2007:RegistryStoredQuery");

            AdhocQueryResponse registryResponseType = fromXml(responseXml);
            return registryResponseType;
        } catch (JAXBException|SOAPException|IOException e) {
            throw new DDSTemporaryException("Unexpected DDS error " + e.getMessage(), e);
        }
    }

    public String getRegistryIdByFSKDocumentId(String documentId) throws DDSPermanentException, DDSTemporaryException {
        AdhocQueryResponse response = getDocumentById(documentId);
        if (!STATUS_SUCCESS.equals(response.getStatus())) {
            throw new DDSPermanentException("Unexpected status from DDS: " + response.getStatus());
        }
        if (response != null &&
                response.getRegistryObjectList() != null &&
                response.getRegistryObjectList().getIdentifiable() != null &&
                response.getRegistryObjectList().getIdentifiable().size() > 0 &&
                response.getRegistryObjectList().getIdentifiable().get(0) != null &&
                response.getRegistryObjectList().getIdentifiable().get(0).getValue() != null) {
            if (response.getRegistryObjectList().getIdentifiable().size() != 1) {
                throw new DDSPermanentException("Unexpected DDS result size for documentId=" + documentId + ". Expected 1, was: " +  response.getRegistryObjectList().getIdentifiable().size());
            }
            List<JAXBElement<? extends IdentifiableType>> identifiable = response.getRegistryObjectList().getIdentifiable();
            IdentifiableType value = identifiable.get(0).getValue();
            String id = value.getId();
            return id;
        } else {
            return null;
        }
    }

    private AdhocQueryResponse fromXml(String responseXml) throws DDSTemporaryException, IOException, SOAPException, JAXBException {
        SOAPBody body = getSoapBody(responseXml);
        if (body.getFault() != null) {
            throw new DDSTemporaryException("Error calling DDS: " + body.getFault().getFaultString());
        }

        JAXBElement<AdhocQueryResponse> jaxbResponse = getUnmarshaller(AdhocQueryResponse.class).unmarshal(body.getFirstChild(), AdhocQueryResponse.class);
        return jaxbResponse.getValue();
    }

}
