/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.dds;

import dk.cprsubscriber.dao.PersonSimplifiedDAO;
import dk.cprsubscriber.domain.PersonSimplifiedVO;
import dk.cprsubscriber.exceptions.CPRDAOException;
import dk.stamkort.dao.PropertiesDAO;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.dao.exceptions.DDSPermanentException;
import dk.stamkort.dao.exceptions.DDSTemporaryException;
import dk.stamkort.domain.PropertiesVO;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.health.LastStateHealthReporter;
import dk.stamkort.util.Properties;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.text.ParseException;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.List;
import java.util.UUID;

@Component
@EnableScheduling
public class SyncJob implements LastStateHealthReporter {
    private static final Logger log = LogManager.getLogger(SyncJob.class);

    public static final String LAST_CPR_SYNC_TIME = "LAST_CPR_SYNC_TIME";
    public static final String LAST_CPR_MAX_PID = "LAST_CPR_MAX_PID";
    public static final String STATUS_DECEASED = "90";
    static String PROPERTIES_DATE_FORMAT_PATTERN = "yyyy-MM-dd HH:mm:ss";
    public static DateTimeFormatter PROPERTIES_DATE_FORMAT = DateTimeFormatter.ofPattern(PROPERTIES_DATE_FORMAT_PATTERN);

    @Autowired
    Properties properties;

    @Autowired
    PersonSimplifiedDAO personSimplifiedDAO;

    @Autowired
    RegistryIndexDAO registryIndexDAO;

    @Autowired
    PropertiesDAO propertiesDAO;

    @Autowired
    DDSRegisterOnDemandClient ddsRegister;

    @Autowired
    DDSRegistryStoredQueryClient ddsQuery;

    @Autowired
    DDSRegistryUpdateClient ddsUpdate;

    private Boolean lastExecutionSucceeded;
    private Instant timeOfLastExecution;
    private Exception lastException;

    public void runSyncPersonIds() {

        debug("SyncJob starting");
        JobStatus jobStatus = new JobStatus(properties.getJobsDDSyncMaxErrors());
        try {
            Instant latestRuntime = getLatestRuntime();
            debug("Latest SyncJob run time: " + latestRuntime);
            long latestMaxPid = getLatestMaxPid();
            Instant jobStartTime = Instant.now();
            Instant jobStartTimeAYearAgo = jobStartTime.atZone(ZoneId.systemDefault()).minusYears(1).toInstant();
            List<PersonSimplifiedVO> updatedPersons = getPersonsUpdatedBetween(latestRuntime, jobStartTime, latestMaxPid);

            int loops = 0;
            long maxPid= 0;

            jobLoop:
            while (updatedPersons.size() > 0 && loops++ < properties.getJobsDDSyncCprMaxLoops()) {
                long maxBatchPid = 0;

                personLoop:
                for (PersonSimplifiedVO personRecord : updatedPersons) {
                    if (personRecord == null) {
                        continue;
                    }
                    if (personRecord.getPid() > maxBatchPid) {
                        maxBatchPid = personRecord.getPid();
                    }

                    boolean personExistsInIndex = personExistsInIndex(personRecord);
                    String statusOneYearAgo = getStatusAtTime(personRecord, jobStartTimeAYearAgo);
                    debug("Person " + personRecord.getCpr() + " with status " + personRecord.getStatus() + " updated. ExistsInIndex=" + personExistsInIndex + ", StatusOneYearAgo=" + statusOneYearAgo);
                    if (!statusIsDeceased(personRecord)) {
                        if (!personExistsInIndex && !statusIsDeceased(statusOneYearAgo)) {
                            // If the person was alive one year ago, the metadata must be available (regardless of current status)
                            String fskDocumentId = createFSKDocumentId();
                            try {
                                createDDRegistryEntry(personRecord, fskDocumentId);
                            } catch (DDSPermanentException e) {
                                log.warn("DDS logical error for documentId=" + fskDocumentId, e);
                                jobStatus.increaseNumberOfErrors();
                                if (jobStatus.tooManyErrors()) {
                                    log.error(jobStatus.errorText());
                                    break jobLoop;
                                } else {
                                    log.warn(jobStatus.warnText());
                                    // Ignore record, continue with next
                                    continue personLoop;
                                }
                            }
                            addPersonToIndex(personRecord, fskDocumentId);
                        }
                    } else {
                        // Person is deceased
                        if (personExistsInIndex && !personIsMarkedForDeletion(personRecord)) {
                            markPersonForDeletion(personRecord);
                        }
                    }
                }
                maxPid = maxBatchPid;
                updatedPersons = getPersonsUpdatedBetween(latestRuntime, jobStartTime, maxBatchPid);
            }

            cleanupPersonsToBeDeleted(jobStatus, jobStartTime);

            writeLatestRuntime(jobStartTime);
            writeHighestPid(maxPid);

            lastExecutionSucceeded = true;
            timeOfLastExecution = Instant.now();
        } catch (DDSTemporaryException e) {
            log.warn("CPR -> DDS synchronization job stopped due to temporary error. Sync will pause and retry later", e);

            lastExecutionSucceeded = false;
            timeOfLastExecution = Instant.now();
            lastException = e;
        } catch (CPRDAOException | DAOException | ParseException e) {
            log.error("CPR -> DDS synchronization job stopped due to internal error", e);

            lastExecutionSucceeded = false;
            timeOfLastExecution = Instant.now();
            lastException = e;
        }
        debug("SyncJob done");
    }

    private void addPersonToIndex(PersonSimplifiedVO personRecord, String fskDocumentId) throws DAOException {
        try {
            RegistryIndexVO index = new RegistryIndexVO();
            index.setFSKDocumentId(fskDocumentId);
            index.setPersonIdentifier(personRecord.getCpr());
            index.setHashValue(""+ personRecord.hashCode());
            debug("Adding person to index: " + personRecord.getCpr());
            registryIndexDAO.createRegistryIndexVO(index);
        } catch (RuntimeException e) {
            // We created DDS metadata, but were unable to write it to our index. The consequence is that
            // the job will add the metadata to DDS again, thereby storing two metadata records. Therefore, we
            // need to make sure that the first metadata is manually deprecated. This is logged:
            log.error("Failed to write to RegistryIndex. This document must be manually deprecated in DDS: documentId="
                    + fskDocumentId + ", cpr=" + personRecord.getCpr());
            throw e;
        }
    }

    private String getStatusAtTime(PersonSimplifiedVO person, Instant time) throws CPRDAOException {
        PersonSimplifiedVO currentPerson = personSimplifiedDAO.getRecordByDate(person.getCpr(), time);
        return (currentPerson != null ? currentPerson.getStatus() : null);
    }

    private String createFSKDocumentId() {
        return "urn:sds:fsk:stamkort:" + UUID.randomUUID().toString();
    }

    private void cleanupPersonsToBeDeleted(JobStatus jobStatus, Instant jobStartTime) throws DAOException {
    	
        List<RegistryIndexVO> fskDocumentIds = registryIndexDAO.getFSKDocumentIdsToBeDeleted(jobStartTime);
        debug("Now deleting: " + fskDocumentIds.size() + " records older than " + jobStartTime);
        for (RegistryIndexVO registryIndex : fskDocumentIds) {
            String fskDocumentId = registryIndex.getFSKDocumentId();
            try {
                String ddsId = ddsQuery.getRegistryIdByFSKDocumentId(fskDocumentId);
                ddsUpdate.deprecateDocumentUsingDdsId(registryIndex.getPersonIdentifier(), ddsId);
                registryIndexDAO.deleteByDocumentId(fskDocumentId);
            } catch (DDSTemporaryException e) {
                log.warn("Now pausing SyncJob: Unable to deprecate document in registry (temporary error), fSKDocumentId=" + fskDocumentId, e);
                break;
            } catch (DDSPermanentException e) {
                jobStatus.increaseNumberOfErrors();
                log.warn("Unable to deprecate document in registry (permanent error), fSKDocumentId=" + fskDocumentId, e);
                if (jobStatus.tooManyErrors()) {
                    log.error(jobStatus.errorText());
                } else {
                    log.warn(jobStatus.warnText());
                }
            }
        }
    }

    private void markPersonForDeletion(PersonSimplifiedVO person) throws DAOException {
        debug("Marking person for deletion: " + person.getCpr());
        registryIndexDAO.markForDeletion(person.getCpr(), oneYearFromNow());
    }

    private Instant oneYearFromNow() {
    	return ZonedDateTime.now().plusYears(1).toInstant();
    }

    private boolean personIsMarkedForDeletion(PersonSimplifiedVO person) throws DAOException {
        RegistryIndexVO registryIndexVO = registryIndexDAO.getRegistryIndexVO(person.getCpr());
        return registryIndexVO.getDeleteAfter() != null;
    }

    private void createDDRegistryEntry(PersonSimplifiedVO person, String fskDocumentId) throws DDSPermanentException, DDSTemporaryException {
        ddsRegister.registerOnDemandDocument(fskDocumentId, person.getCpr());
    }

    private boolean personExistsInIndex(PersonSimplifiedVO person) throws DAOException {
        RegistryIndexVO registryIndexVO = registryIndexDAO.getRegistryIndexVO(person.getCpr());
        return registryIndexVO != null;
    }

    private boolean statusIsDeceased(PersonSimplifiedVO person) {
        return statusIsDeceased(person.getStatus());
    }

    private boolean statusIsDeceased(String status) {
        return STATUS_DECEASED.equals(status);
    }

    private List<PersonSimplifiedVO> getPersonsUpdatedBetween(Instant latestRuntime, Instant startTime, long minPid) throws CPRDAOException {
        List<PersonSimplifiedVO> updated = personSimplifiedDAO.getPersonsUpdatedBetween(latestRuntime, startTime, minPid, properties.getJobsDDSyncCprMaxResults());
        debug("getPersonsUpdatedBetween(" + latestRuntime + ", " + startTime + ", " + minPid + ") -> " + updated.size() + " records");
        return updated;
    }

    private void debug(String s) {
        log.debug(s);
    }

    private Instant getLatestRuntime() throws ParseException, DAOException {
        PropertiesVO lastDeletionTime = propertiesDAO.getProperty(LAST_CPR_SYNC_TIME);
        if (lastDeletionTime == null || lastDeletionTime.getPropertyValue() == null) {
            throw new DAOException("Last CPR synchronization not found in database (propertyKey " + LAST_CPR_SYNC_TIME + " with format " + PROPERTIES_DATE_FORMAT_PATTERN + ")");
        }
        LocalDateTime ldt = LocalDateTime.parse(lastDeletionTime.getPropertyValue(), PROPERTIES_DATE_FORMAT);
   	  	ZonedDateTime zdt = ldt.atZone(ZoneId.systemDefault());
        Instant latestRuntime = zdt.toInstant();
        return latestRuntime;

    }

    private long getLatestMaxPid() throws DAOException {
        PropertiesVO lastCprMaxPid = propertiesDAO.getProperty(LAST_CPR_MAX_PID);
        if (lastCprMaxPid == null || lastCprMaxPid.getPropertyValue() == null) {
            throw new DAOException("Max CPR PID not found in database (propertyKey " + LAST_CPR_MAX_PID + ")");
        }
        try {
            Long.valueOf(lastCprMaxPid.getPropertyValue());
        } catch (NumberFormatException e) {
            throw new DAOException("Unparsable long value from database (propertyKey " + LAST_CPR_MAX_PID + ", value=" + lastCprMaxPid.getPropertyValue() + ")");
        }
        return Long.valueOf(lastCprMaxPid.getPropertyValue());
    }

    private void writeLatestRuntime(Instant jobStartTime) throws DAOException {
    	ZonedDateTime zdt = jobStartTime.atZone(ZoneId.systemDefault());
    	String propDate = PROPERTIES_DATE_FORMAT.format(zdt);
    	propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDate);
    }

    private void writeHighestPid(long maxPid) throws DAOException {
        propertiesDAO.updateProperty(LAST_CPR_MAX_PID, "" + maxPid);
    }

    @Override
    public Boolean lastExecutionSucceeded() {
        return lastExecutionSucceeded;
    }

    @Override
    public Instant timeOfLastExecution() {
        return timeOfLastExecution;
    }

    @Override
    public Exception lastException() {
        return lastException;
    }
}
