/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.dds;

import dk.stamkort.OID;
import oasis.names.tc.ebxml_regrep.xsd.lcm._3.ObjectFactory;
import oasis.names.tc.ebxml_regrep.xsd.lcm._3.SubmitObjectsRequest;
import oasis.names.tc.ebxml_regrep.xsd.rim._3.*;

import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.UUID;

import static dk.stamkort.integrations.dds.XdsDocumentEntryConstants.*;

public class XDSRequestHelper {

    // FSK-specific XDS constants
    private static final String FSK_NAME = "Fælles stamkort";
    private static final String DK_FSK_SCHEMA = "DK FSK Schema";
    private static final String FSK_LANGUAGE_CODE = "da-DK";
    private static final String FSK_CODING_SCHEME_URN = "urn:ad:dk:medcom:fsk:full";
    private static final String CONFIDENTIALITY_NAME = "Normal";
    private static final String SOURCE_OBJECT_SUBMISSION_SET = "SubmissionSet";
    private static final String ASSOCIATION_ORIGINAL_STATUS = "OriginalStatus";
    private static final String ASSOCIATION_NEW_STATUS = "NewStatus";
    private static final String MIME_TYPE = "text/xml";
    private static final String CODING_SCHEME_LOINC = "LOINC";
    private static final String CLASSIFICATION_NODE_REPRESENTATION = "001";
    private static final String TYPECODE_NAME_PATIENT_INFORMATION = "Patient Information";

    private final DateTimeFormatter submissionTimeSDF = DateTimeFormatter.ofPattern("yyyyMMddHHmmss").withZone(ZoneId.systemDefault());

    public String getPatientId(String patientCPR) {
        return patientCPR + "^^^&" + OID.CPR + "&ISO";
    }

    private oasis.names.tc.ebxml_regrep.xsd.rim._3.ObjectFactory objFactory = new oasis.names.tc.ebxml_regrep.xsd.rim._3.ObjectFactory();

    public SubmitObjectsRequest create() {
        ObjectFactory objFactory = new ObjectFactory();
        return objFactory.createSubmitObjectsRequest();
    }

    public ExtrinsicObjectType addOnDemandDocumentMetadataEntry(SubmitObjectsRequest request, String documentId, String repositoryUniqueId, String homeCommunityId) {
        ExtrinsicObjectType onDemandDocumentMetadata = createOnDemandDocumentMetadataEntry(UUID.randomUUID().toString(), repositoryUniqueId);
        onDemandDocumentMetadata.setHome(homeCommunityId);
        onDemandDocumentMetadata.setStatus(XDS_DOCUMENTENTRY_STATUS_APPROVED);
        if (request.getRegistryObjectList() == null) {
            request.setRegistryObjectList(objFactory.createRegistryObjectListType());
        }
        request.getRegistryObjectList().getIdentifiable().add(objFactory.createExtrinsicObject(onDemandDocumentMetadata));

        ExternalIdentifierType uniqueIdIdentifier = createUniqueIdExternalIdentifierType(documentId, onDemandDocumentMetadata);
        onDemandDocumentMetadata.getExternalIdentifier().add(uniqueIdIdentifier);
        InternationalStringType name = new InternationalStringType();
        LocalizedStringType nameStr = new LocalizedStringType();
        nameStr.setValue(FSK_NAME);
        name.getLocalizedString().add(nameStr);
        onDemandDocumentMetadata.setName(name);

        return onDemandDocumentMetadata;
    }

    public void addDocumentMetadataAttributes(ExtrinsicObjectType documentMetadataEntry, String authorInstitution, String patientId, String typeCode) {
        ClassificationType classification = objFactory.createClassificationType();
        classification.setId(UUID.randomUUID().toString());
        classification.setClassificationScheme(XDS_DOCUMENTENTRY_CLASSIFICATION_SCHEME_CLASSIFICATION_AUTHOR);
        classification.setClassifiedObject(documentMetadataEntry.getId());
        classification.setNodeRepresentation(CLASSIFICATION_NODE_REPRESENTATION);
        addFormatCodeClassification(documentMetadataEntry);
        addTypeCodeClassification(documentMetadataEntry, typeCode);
        addSlot(classification, XDS_DOCUMENTENTRY_SLOT_NAME_AUTHOR_INSTITUTION, authorInstitution);
        classification.setName(objFactory.createInternationalStringType());
        List<LocalizedStringType> localizedString = classification.getName().getLocalizedString();
        LocalizedStringType e = new LocalizedStringType();
        e.setValue(CONFIDENTIALITY_NAME);
        localizedString.add(e);
        addSlot(classification, XDS_SLOT_NAME_CODING_SCHEME, OID.CONFIDENTIALITY_CODING_SCHEME_OID);
        documentMetadataEntry.getClassification().add(classification);
        addSlot(documentMetadataEntry, XDS_DOCUMENTENTRY_SLOT_NAME_LANGUAGE_CODE, FSK_LANGUAGE_CODE);
        addSlot(documentMetadataEntry, XDS_DOCUMENTENTRY_SLOT_NAME_SOURCE_PATIENT_ID, patientId);
        ExternalIdentifierType patientIdIdentifier = objFactory.createExternalIdentifierType();
        patientIdIdentifier.setId(UUID.randomUUID().toString());
        patientIdIdentifier.setIdentificationScheme(XDS_DOCUMENTENTRY_IDENTIFICATION_SCHEME_EXTERNAL_IDENTIFIER_PATIENT_ID);
        patientIdIdentifier.setRegistryObject(documentMetadataEntry.getId());
        patientIdIdentifier.setValue(patientId);
        documentMetadataEntry.getExternalIdentifier().add(patientIdIdentifier);
    }

    private void addFormatCodeClassification(ExtrinsicObjectType documentMetadataEntry) {
        ClassificationType classification = objFactory.createClassificationType();
        classification.setClassificationScheme(XDS_DOCUMENTENTRY_CLASSIFICATION_SCHEME_CLASSIFICATION_FORMAT_CODE);
        classification.setClassifiedObject(documentMetadataEntry.getId());
        classification.setId(UUID.randomUUID().toString());
        classification.setNodeRepresentation(DK_FSK_SCHEMA);
        LocalizedStringType name = new LocalizedStringType();
        name.setValue(DK_FSK_SCHEMA);
        InternationalStringType nameObj = new InternationalStringType();
        nameObj.getLocalizedString().add(name);
        classification.setName(nameObj);
        addSlot(classification, XDS_SLOT_NAME_CODING_SCHEME, FSK_CODING_SCHEME_URN);
        documentMetadataEntry.getClassification().add(classification);
    }

    private void addTypeCodeClassification(ExtrinsicObjectType documentMetadataEntry, String typeCode) {
        ClassificationType classification = objFactory.createClassificationType();
        classification.setClassificationScheme(XDS_DOCUMENTENTRY_CLASSIFICATION_SCHEME_CLASSIFICATION_HEALTHCARE_FACILITY_TYPE_CODE);
        classification.setClassifiedObject(documentMetadataEntry.getId());
        classification.setId(UUID.randomUUID().toString());
        classification.setNodeRepresentation(typeCode);
        LocalizedStringType name = new LocalizedStringType();
        name.setValue(TYPECODE_NAME_PATIENT_INFORMATION);
        InternationalStringType nameObj = new InternationalStringType();
        nameObj.getLocalizedString().add(name);
        classification.setName(nameObj);
        addSlot(classification, XDS_SLOT_NAME_CODING_SCHEME, CODING_SCHEME_LOINC);
        documentMetadataEntry.getClassification().add(classification);
    }

    private void addSlot(IdentifiableType classification, String slotName, String slotValue) {
        SlotType1 slot = objFactory.createSlotType1();
        slot.setName(slotName);
        slot.setValueList(objFactory.createValueListType());
        slot.getValueList().getValue().add(slotValue);
        classification.getSlot().add(slot);
    }

    public RegistryPackageType addSubmissionSet(SubmitObjectsRequest request, String homeCommunityId, String patientId, String documentId) {
        RegistryPackageType registryPackage = objFactory.createRegistryPackageType();
        registryPackage.setId(UUID.randomUUID().toString());
        registryPackage.setObjectType(XSD_DOCUMENTENTY_REGISTRY_PACKAGE_OBJECTTYPE);
        registryPackage.setHome(homeCommunityId);

        addSlot(registryPackage, XDS_SLOT_NAME_SUBMISSION_TIME, submissionTimeSDF.format(Instant.now()));
        addExternalIdentifier(registryPackage, XDS_EXTERNAL_ID_UNIQUE_ID, XDS_SUBMISSIONSET_UNIQUE_ID, documentId);
        addExternalIdentifier(registryPackage, XDS_EXTERNAL_ID_PATIENT_ID, XDS_SUBMISSIONSET_PATIENT_ID, patientId);

        request.getRegistryObjectList().getIdentifiable().add(objFactory.createRegistryPackage(registryPackage));

        // Classify registry package as XDSSubmissionSet (outside the RegistryPackage)
        ClassificationType classification = objFactory.createClassificationType();
        classification.setClassificationNode(XDS_SUBMISSTION_SET);
        classification.setId(UUID.randomUUID().toString());
        classification.setObjectType(XDS_OBJECTTYPE_CLASSIFICATION);
        classification.setClassifiedObject(registryPackage.getId());
        request.getRegistryObjectList().getIdentifiable().add(objFactory.createClassification(classification));

        return registryPackage;
    }

    public RegistryPackageType addSubmissionSetForUpdate(SubmitObjectsRequest request, String homeCommunityId, String patientId, String documentId, String documentSource, String typeCode) {
        request.setRegistryObjectList(objFactory.createRegistryObjectListType());
        RegistryPackageType registryPackage = objFactory.createRegistryPackageType();
        registryPackage.setId(SOURCE_OBJECT_SUBMISSION_SET);
        registryPackage.setObjectType(XSD_DOCUMENTENTY_REGISTRY_PACKAGE_OBJECTTYPE);
        registryPackage.setHome(homeCommunityId);
        request.getRegistryObjectList().getIdentifiable().add(objFactory.createRegistryPackage(registryPackage));

        ClassificationType classification = objFactory.createClassificationType();
        classification.setClassificationNode(XDS_SUBMISSTION_SET);
        classification.setClassifiedObject(SOURCE_OBJECT_SUBMISSION_SET);
        registryPackage.getClassification().add(classification);

        ClassificationType codingScheme = objFactory.createClassificationType();
        codingScheme.setClassificationScheme(XDS_CONTENT_TYPE);
        codingScheme.setClassifiedObject(SOURCE_OBJECT_SUBMISSION_SET);
        codingScheme.setNodeRepresentation(typeCode);
        classification.setObjectType(XDS_OBJECTTYPE_CLASSIFICATION);
        addSlot(codingScheme, XDS_SLOT_NAME_CODING_SCHEME, OID.FSK_CODING_SCHEME_OID);
        InternationalStringType name = new InternationalStringType();
        LocalizedStringType classificationName = new LocalizedStringType();
        classificationName.setValue(FSK_NAME);
        name.getLocalizedString().add(classificationName);
        codingScheme.setName(name);
        registryPackage.getClassification().add(codingScheme);

        addExternalIdentifier(registryPackage, XDS_EXTERNAL_ID_UNIQUE_ID, null, UUID.randomUUID().toString());
        addExternalIdentifier(registryPackage, XDS_EXTERNAL_ID_PATIENT_ID, null, patientId);
        addExternalIdentifier(registryPackage, XDS_DOCUMENTENTRY_IDENTIFICATION_SCHEME_EXTERNAL_IDENTIFIER_UNIQUE_ID, null, documentId);
        addExternalIdentifier(registryPackage, XDS_EXTERNAL_ID_DOCUMENT_SOURCE, null, documentSource);

        addSlot(registryPackage, XDS_SLOT_NAME_SUBMISSION_TIME, submissionTimeSDF.format(Instant.now()));

        return registryPackage;
    }

    private void addExternalIdentifier(RegistryPackageType registryPackage, String identificationScheme, String localizedString, String value) {
        ExternalIdentifierType externalId = new ExternalIdentifierType();
        externalId.setIdentificationScheme(identificationScheme);
        externalId.setValue(value);
        externalId.setId(UUID.randomUUID().toString());
        InternationalStringType name = new InternationalStringType();
        if (localizedString != null) {
            LocalizedStringType localizedStringObj = new LocalizedStringType();
            localizedStringObj.setValue(localizedString);
            name.getLocalizedString().add(localizedStringObj);
        }
        externalId.setName(name);
        registryPackage.getExternalIdentifier().add(externalId);
    }

    public AssociationType1 addAssociationForUpdate(SubmitObjectsRequest request, String ddsDocumentId) {
        AssociationType1 association = objFactory.createAssociationType1();
        association.setAssociationType(XSD_UPDATE_AVAILABILITY_STATUS);
        association.setSourceObject(SOURCE_OBJECT_SUBMISSION_SET);
        association.setTargetObject(ddsDocumentId);
        association.setId(UUID.randomUUID().toString());

        addSlot(association, ASSOCIATION_ORIGINAL_STATUS, XDS_DOCUMENTENTRY_STATUS_APPROVED);
        //addSlot(association, ASSOCIATION_NEW_STATUS, XDS_DOCUMENTENTRY_STATUS_APPROVED);
        addSlot(association, ASSOCIATION_NEW_STATUS, XDS_DOCUMENTENTRY_STATUS_DEPRECATED);

        request.getRegistryObjectList().getIdentifiable().add(objFactory.createAssociation(association));
        return association;
    }

    private ExternalIdentifierType createUniqueIdExternalIdentifierType(String documentId, ExtrinsicObjectType metadataObjId) {
        ExternalIdentifierType uniqueIdIdentifier = objFactory.createExternalIdentifierType();
        uniqueIdIdentifier.setId(UUID.randomUUID().toString());
        uniqueIdIdentifier.setIdentificationScheme(XDS_DOCUMENTENTRY_IDENTIFICATION_SCHEME_EXTERNAL_IDENTIFIER_UNIQUE_ID);
        uniqueIdIdentifier.setRegistryObject(metadataObjId.getId());
        uniqueIdIdentifier.setName(objFactory.createInternationalStringType());
        LocalizedStringType name = objFactory.createLocalizedStringType();
        name.setValue(XDS_DOCUMENTENTRY_UNIQUE_ID);
        uniqueIdIdentifier.getName().getLocalizedString().add(name);
        uniqueIdIdentifier.setValue(documentId);
        return uniqueIdIdentifier;
    }

    private ExtrinsicObjectType createOnDemandDocumentMetadataEntry(String documentId, String repositoryUniqueId) {
        return createDocumentMetadataEntry(documentId, XDS_DOCUMENTENTRY_OBJ_TYPE_ONDEMAND_DOCUMENT, repositoryUniqueId);
    }

    private ExtrinsicObjectType createDocumentMetadataEntry(String documentId, String documentObjType, String repositoryUniqueId) {
        ExtrinsicObjectType extrinsicObjectType = objFactory.createExtrinsicObjectType();
        extrinsicObjectType.setId(documentId);
        extrinsicObjectType.setMimeType(MIME_TYPE);
        extrinsicObjectType.setObjectType(documentObjType);

        addSlot(extrinsicObjectType, XDS_DOCUMENTENTRY_SLOT_NAME_REPOSITORY_UNIQUE_ID, repositoryUniqueId);
        return extrinsicObjectType;
    }

}
