/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda;

import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.AuthorProperties;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.util.Properties;
import dk.stamkort.webservice.hl7cda.assembler.PDC_1_0ClinicalDocumentAssembler;
import dk.stamkort.webservice.hl7cda.assembler.PDC_2_0ClinicalDocumentAssembler;
import dk.stamkort.webservice.hl7cda.model.PDC_2_0DataHolder;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import oio.medcom.cprservice._1_0.PersonCivilRegistrationStatusStructureType;
import oio.medcom.cprservice._1_0.PersonHealthCareInformationStructureType;
import org.hl7.v3.POCDMT000040ClinicalDocument;
import org.springframework.stereotype.Service;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.namespace.QName;
import java.io.StringWriter;
import java.math.BigInteger;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;

@Service
public class ClinicalDocumentProviderImpl implements ClinicalDocumentProvider {

    private final PersonalDataCardServiceImpl personalDataCardService;
    private final Properties properties;
    private final Marshaller marshaller;

    public ClinicalDocumentProviderImpl(PersonalDataCardServiceImpl personalDataCardService, Properties properties) throws JAXBException {
        this.personalDataCardService = personalDataCardService;
        this.properties = properties;

        marshaller = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class).createMarshaller();
    }

    @Override
    public String getClinicalDocument(String cprnr, String typeCode, String formatCode) throws JAXBException {
        RegisterData registerData = personalDataCardService.getRegisterData(cprnr);
        checkStatus(getStatus(registerData));

        GetPersonWithHealthCareInformationOut scesData = registerData.getScesData();
        PersonHealthCareInformationStructureType healthCareInformation = scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure();

        String yderNr = getYderNr(healthCareInformation);

        String document;
        if (Metadata.PDC_2_0_TYPE_CODE.equals(typeCode) && Metadata.PDC_2_0_FORMAT_CODE.equals(formatCode)) {
            document = createPDC_2_0ClinicalDocument(registerData, registerData.getYder(yderNr));
        } else {
            // Fall back to old format
            document = createPDC_1_0ClinicalDocument(registerData, yderNr);
        }

        return document;
    }

    private String createPDC_1_0ClinicalDocument(RegisterData registerData, String yderNr) throws JAXBException {
        POCDMT000040ClinicalDocument clinicalDocument = PDC_1_0ClinicalDocumentAssembler.assembleClinicalDocument(registerData, yderNr);
        JAXBElement<POCDMT000040ClinicalDocument> documentJAXBElement = new JAXBElement<>(new QName("urn:hl7-org:v3", "POCD_MT000040.ClinicalDocument"), POCDMT000040ClinicalDocument.class, clinicalDocument);

        StringWriter sw = new StringWriter();
        marshaller.marshal(documentJAXBElement, sw);
        return sw.toString();
    }

    private String createPDC_2_0ClinicalDocument(RegisterData registerData, YderVO yder) throws JAXBException {
        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        AuthorProperties authorProperties = properties.createAuthorProperties();
        POCDMT000040ClinicalDocument clinicalDocument = PDC_2_0ClinicalDocumentAssembler.assembleClinicalDocument(dataHolder, authorProperties);
        JAXBElement<POCDMT000040ClinicalDocument> documentJAXBElement = new JAXBElement<>(new QName("urn:hl7-org:v3", "ClinicalDocument"), POCDMT000040ClinicalDocument.class, clinicalDocument);

        StringWriter sw = new StringWriter();
        marshaller.marshal(documentJAXBElement, sw);
        return enhanceXml(sw.toString());
    }

    private String enhanceXml(String toString) {
        return new XmlEnhancer().enhance(toString);
    }

    private void checkStatus(PersonCivilRegistrationStatusStructureType statusStructure) {
        if (statusStructure == null) {
            return;
        }
        String status = statusStructure.getPersonCivilRegistrationStatusCode();
        switch (status) {
            case "30":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_30, properties.getCallScesStatus30());
                break;
            case "50":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_50, properties.getCallScesStatus50());
                break;
            case "60":
                throwFaultStatus(FaultCode.CALL_SCES_STATUS_60, properties.getCallScesStatus60());
                break;
            case "90":
                XMLGregorianCalendar statusDate = statusStructure.getPersonCivilRegistrationStatusStartDate();
                LocalDateTime statusDateTime = statusDate.toGregorianCalendar().toZonedDateTime().toLocalDateTime();
                LocalDateTime oneYear = LocalDateTime.now().minus(1, ChronoUnit.YEARS);
                if (statusDateTime.isBefore(oneYear)) {
                    throwFaultStatus(FaultCode.CALL_SCES_STATUS_90, properties.getCallScesStatus90());
                }

                break;
        }

    }

    private PersonCivilRegistrationStatusStructureType getStatus(RegisterData registerData) {
        return registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().getPersonCivilRegistrationStatusStructure();
    }

    private void throwFaultStatus(FaultCode faultCode, String faultText) {
        if (org.apache.commons.lang.StringUtils.isEmpty(faultText)) {
            FaultUtil.throwFSKFault(faultCode);
        } else {
            FaultUtil.throwFSKFault(faultCode, faultText);
        }
    }

    private String getYderNr(PersonHealthCareInformationStructureType healthCareInformation) {
        String yderNr = "";
        if (healthCareInformation != null) {
            AssociatedGeneralPractitionerStructureType associatedGeneralPractitioner = healthCareInformation.getCurrentGeneralPractitionerStructure().getAssociatedGeneralPractitionerStructure();
            if (associatedGeneralPractitioner != null) {
                BigInteger ydernrBI = associatedGeneralPractitioner.getAssociatedGeneralPractitionerIdentifier();
                if (ydernrBI != null && !ydernrBI.equals(new BigInteger("0"))) {
                    yderNr = ydernrBI.toString();
                }
            }
        }

        return yderNr;
    }
}
