/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda;

import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.CountryIdentificationCodeType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.webservice.hl7cda.model.*;
import dk.sundhedsdatastyrelsen.fsk._2019._08._01.GetPersonalDataCardResponse;
import dk.sundhedsdatastyrelsen.fsk.extensions.*;
import oio.medcom.cprservice._1_0.*;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.hl7.v3.*;

import javax.xml.bind.JAXBElement;
import javax.xml.datatype.XMLGregorianCalendar;
import java.io.Serializable;
import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * The idea behind this data mapper is to populate the models at best effort by using data from the data sources.
 * Then the CDA-assembler/mapper classes will determine if something required is missing, and then fail.
 */
public class PDCDataMapper {

    private static final DateTimeFormatter dateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMddHHmmssZ"); // Output sample: 20171019144200+0200
    private static final DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd");

    public static PDC_2_0DataHolder mapDataHolder(RegisterData registerData, YderVO yder) {
        PrimaryPersonInformationHolder personHolder = mapPersonHolder(registerData);
        ProviderOrganizationHolder providerOrganizationHolder = mapProviderOrganizationHolder(yder);
        List<CustodyHolder> custodyHolder = mapCustodyHolder(registerData);
        CoverageGroupHolder coverageGroupHolder = mapCoverageGroupHolder(registerData);
        OrganDonorRegistrationHolder organDonorRegistrationHolder = mapOrganDonorRegistrationHolder(registerData);
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = mapTreatmentWillRegistrationHolder(registerData);
        LivingWillRegistrationHolder livingWillRegistrationHolder = mapLivingWillRegistrationHolder(registerData);
        List<LanguageTypedInHolder> languageTypedInHolder = mapLanguageTypedInHolder(registerData);
        List<TempAddrTypedInHolder> tempAddrTypedInHolder = mapTempAddrTypedInHolder(registerData);
        List<DentistTypedInHolder> dentistTypedInHolder = mapDentistTypedInHolder(registerData);
        List<PatientContactTypedInHolder> patientContactTypedInHolder = mapPatientContactTypedInHolder(registerData);
        List<RelativeTypedInHolder> relativeTypedInHolder = mapRelativeTypedInHolder(registerData);

        return new PDC_2_0DataHolder()
                .setPrimaryPersonInformation(personHolder)
                .setProviderOrganizationHolder(providerOrganizationHolder)
                .setCustodyHolderList(custodyHolder)
                .setCoverageGroupHolder(coverageGroupHolder)
                .setOrganDonorRegistrationHolder(organDonorRegistrationHolder)
                .setTreatmentWillRegistrationHolder(treatmentWillRegistrationHolder)
                .setLivingWillRegistrationHolder(livingWillRegistrationHolder)
                .setLanguageTypedInHolder(languageTypedInHolder)
                .setTempAddrTypedInHolder(tempAddrTypedInHolder)
                .setDentistTypedInHolder(dentistTypedInHolder)
                .setPatientContactTypedInHolder(patientContactTypedInHolder)
                .setRelativeTypedInHolder(relativeTypedInHolder);
    }

    private static List<CustodyHolder> mapCustodyHolder(RegisterData registerData) {
        List<CustodyHolder> custodyHolderList = new ArrayList<>();

        // ChildCustody data
        // This custody code should maybe get mapped in the CDA assembler classes instead
        CodeHolder childCustodyCustodyCode = new CodeHolder();
        childCustodyCustodyCode.setCode("ChildCustody");
        childCustodyCustodyCode.setDisplayName("Forældremyndighed over");

        registerData.getChildCustodyCustodyData()
                .stream()
                .map(custodyData ->
                        mapCustodyHolderElement(
                                custodyData.getGetPersonWithHealthCareInformationOut(),
                                childCustodyCustodyCode,
                                mapChildCustodyCustodyRelationCode(
                                        registerData.getCprnr(),
                                        custodyData.getGetPersonWithHealthCareInformationOut())))
                .forEach(custodyHolderList::add);

        // CustodyBy data
        // This custody code should maybe get mapped in the CDA assembler classes instead
        CodeHolder custodyByCustodyCode = new CodeHolder();
        custodyByCustodyCode.setCode("CustodyBy");
        custodyByCustodyCode.setDisplayName("Forældremyndighedshaver");

        registerData.getCustodyByCustodyData()
                .stream()
                .map(custodyData ->
                        mapCustodyHolderElement(
                                custodyData.getGetPersonWithHealthCareInformationOut(),
                                custodyByCustodyCode,
                                mapCustodyByCustodyRelationCode(
                                        custodyData.getPersonCustodyRelationType())))
                .forEach(custodyHolderList::add);

        return custodyHolderList;
    }

    private static CustodyHolder mapCustodyHolderElement(GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformationOut, CodeHolder custodyCode, CodeHolder custodyRelationCode) {
        CustodyHolder custodyHolder = new CustodyHolder();

        // code
        custodyHolder.setCode(custodyCode);

        RegularCPRPersonType regularCPRPerson = getPersonWithHealthCareInformationOut
                .getPersonWithHealthCareInformationStructure() // jaxb required
                .getPersonInformationStructure() // jaxb required
                .getRegularCPRPerson(); // jaxb required

        SimpleCPRPersonType simpleCPRPerson = regularCPRPerson
                .getSimpleCPRPerson(); // jaxb required

        // cpr
        custodyHolder.setCpr(simpleCPRPerson.getPersonCivilRegistrationIdentifier());

        PersonNameStructureType personNameStructure = simpleCPRPerson
                .getPersonNameStructure(); // jaxb optional

        if (personNameStructure != null) {
            // givens
            // Split given names by whitespace
            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonGivenName()))) // jaxb required
                    .forEach(custodyHolder::addGiven);

            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonMiddleName()))) // jaxb optional
                    .forEach(custodyHolder::addGiven); // add middle name to given names

            // family
            custodyHolder.setFamily(
                    personNameStructure.getPersonSurnameName()); // jaxb required
        }

        // relation
        custodyHolder.setRelation(custodyRelationCode);

        return custodyHolder;
    }

    /**
     * @param cprOfParent                                  cpr of the parent
     * @param getPersonWithHealthCareInformationOutOfChild GetPersonWithHealthCareInformationOut object of the Child
     * @return code
     */
    private static CodeHolder mapChildCustodyCustodyRelationCode(String cprOfParent, GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformationOutOfChild) {

        List<PersonCustodyStructureType> personCustodyStructure = getPersonWithHealthCareInformationOutOfChild
                .getPersonWithHealthCareInformationStructure() // jaxb required
                .getPersonInformationStructure() // jaxb required
                .getPersonCustodyStructure();

        // Loop Person Custodies of the child to find the relation to the parent for the current parent
        PersonCustodyRelationType personCustodyRelationType = personCustodyStructure
                .stream()
                .filter(personCustodyStructureType ->
                        personCustodyStructureType.getPersonCustodyIdentifier().equals(cprOfParent))
                .reduce((personCustodyStructure1, personCustodyStructure2) -> {
                    throw new IllegalStateException("Multiple children with same CPR found");
                })
                .map(PersonCustodyStructureType::getPersonCustodyRelation) // jaxb required
                .orElseThrow(() -> new IllegalStateException("Expected Person Custody not found for child"));

        String code;
        String displayName;

        switch (personCustodyRelationType) {
            case MOR:
                code = "mor";
                displayName = "Mor";
                break;
            case FAR:
                code = "far";
                displayName = "Far";
                break;
            case ANDEN:
                code = "anden";
                displayName = "Anden relation";
                break;
            default:
                throw new IllegalStateException("Unexpected value: " + personCustodyRelationType);
        }

        CodeHolder codeHolder = new CodeHolder();

        // code
        codeHolder.setCode(code);

        // displayName
        codeHolder.setDisplayName(displayName);

        return codeHolder;
    }

    @SuppressWarnings("unused") // unused for now, but maybe relevant in future
    private static CodeHolder mapCustodyByCustodyRelationCode(PersonCustodyRelationType personCustodyRelationType) {
        CodeHolder codeHolder = new CodeHolder();

        // According to "Teknisk Implementeringsvejledning" (https://www.nspop.dk/x/YgsaBg)
        // Always use "barn" and "Barn" as values for CustodyBy

        // code
        codeHolder.setCode("barn");

        // displayName
        codeHolder.setDisplayName("Barn");

        return codeHolder;
    }

    private static CoverageGroupHolder mapCoverageGroupHolder(RegisterData registerData) {
        PersonPublicHealthInsuranceType personPublicHealthInsurance = registerData.getScesData()
                .getPersonWithHealthCareInformationStructure() // jaxb required
                .getPersonHealthCareInformationStructure() // jaxb required
                .getPersonPublicHealthInsurance(); // jaxb required

        CoverageGroupHolder coverageGroupHolder = new CoverageGroupHolder();

        String group = Optional.ofNullable(
                personPublicHealthInsurance
                        .getPublicHealthInsuranceGroupIdentifier()) // jaxb required
                .map(PublicHealthInsuranceGroupIdentifierType::value)
                .map(str -> str.replace("sygesikringsgruppe_", "") )
                .orElse(null);

        // group
        coverageGroupHolder.setGroup(group);

        return coverageGroupHolder;
    }

    private static List<RelativeTypedInHolder> mapRelativeTypedInHolder(RegisterData registerData) {
        List<RelativeTypedInHolder> relativeTypedInHolderList = new ArrayList<>();

        getSKREntryElementsOfType(RelatedPerson.class, registerData.getPersonalDataCard())
                .forEach(relatedPerson ->
                        relativeTypedInHolderList.add(mapRelativeTypedInHolderElement(relatedPerson)));

        return relativeTypedInHolderList;
    }

    private static RelativeTypedInHolder mapRelativeTypedInHolderElement(RelatedPerson relatedPerson) {
        RelativeTypedInHolder relativeTypedInHolder = new RelativeTypedInHolder(new IdHolder(relatedPerson.getId().getRoot(), relatedPerson.getId().getExtension(), relatedPerson.getId().getAssigningAuthorityName()));

        POCDMT000040AssociatedEntity associatedEntity = relatedPerson
                .getAssociatedEntity(); // jaxb required

        POCDMT000040Person associatedPerson = associatedEntity.getAssociatedPerson();

        Optional.ofNullable(associatedPerson)
                .ifPresent(value -> {
                    // givens
                    relativeTypedInHolder.setFamily(getPersonFamily(value));

                    // family
                    getPersonGivens(value)
                            .forEach(relativeTypedInHolder::addGiven);
                });

        CD relationshipType = relatedPerson
                .getRelationshipType(); // jaxb required

        CodeHolder relation = new CodeHolder();
        relation.setCode(relationshipType.getCode());
        relation.setDisplayName(relationshipType.getDisplayName());

        // relationCode
        relativeTypedInHolder.setRelation(relation);

        String note = relatedPerson
                .getNote(); // jaxb optional

        // Phone
        associatedEntity.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("tel:"))
                .forEach(tel ->
                        relativeTypedInHolder.getPhoneHolders().add((mapPhoneHolder(tel))));

        // note
        relativeTypedInHolder.setNote(note);

        POCDMT000040Author dataEnterer = relatedPerson
                .getDataEnterer(); // jaxb required

        POCDMT000040AssignedAuthor assignedAuthor = dataEnterer
                .getAssignedAuthor(); // jaxb required

        // time
        relativeTypedInHolder.setAuthorTime(getDataEntererTime(dataEnterer));

        // authorCpr
        relativeTypedInHolder.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor
                .getAssignedPerson(); // jaxb optional

        // authorGivens
        getPersonGivens(assignedPerson)
                .forEach(relativeTypedInHolder::addAuthorGiven);

        // authorFamily
        relativeTypedInHolder.setAuthorFamily(getPersonFamily(assignedPerson));

        return relativeTypedInHolder;
    }

    private static List<PatientContactTypedInHolder> mapPatientContactTypedInHolder(RegisterData registerData) {
        List<PatientContactTypedInHolder> patientContactTypedInHolderList = new ArrayList<>();

        getSKREntryElementsOfType(PatientContact.class, registerData.getPersonalDataCard())
                .forEach(patientContact ->
                        patientContactTypedInHolderList.add(mapPatientContactTypedInHolderElement(patientContact)));

        return patientContactTypedInHolderList;
    }

    public static PatientContactTypedInHolder mapPatientContactTypedInHolderElement(PatientContact patientContact) {
        PatientContactTypedInHolder patientContactTypedInHolder = new PatientContactTypedInHolder();

        // phones
        patientContact.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("tel:"))
                .forEach(tel ->
                        patientContactTypedInHolder.addPhone(mapPhoneHolder(tel)));

        // emailAddresses
        patientContact.getTelecoms()
                .stream()
                .filter(tel -> tel.getValue().startsWith("mailto:"))
                .forEach(tel ->
                        patientContactTypedInHolder.addEmailAddress(mapPhoneHolder(tel)));

        POCDMT000040Author dataEnterer = patientContact
                .getDataEnterer(); // jaxb required

        POCDMT000040AssignedAuthor assignedAuthor = dataEnterer
                .getAssignedAuthor(); // jaxb required

        // time
        patientContactTypedInHolder.setAuthorTime(getDataEntererTime(dataEnterer));

        // authorCpr
        patientContactTypedInHolder.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor
                .getAssignedPerson(); // jaxb optional

        // authorGivens
        getPersonGivens(assignedPerson)
                .forEach(patientContactTypedInHolder::addAuthorGiven);

        // authorFamily
        patientContactTypedInHolder.setAuthorFamily(getPersonFamily(assignedPerson));

        return patientContactTypedInHolder;
    }

    private static List<DentistTypedInHolder> mapDentistTypedInHolder(RegisterData registerData) {
        List<DentistTypedInHolder> dentistTypedInHolderList = new ArrayList<>();

        getSKREntryElementsOfType(HealthProvider.class, registerData.getPersonalDataCard())
                .forEach(healthProvider ->
                        dentistTypedInHolderList.add(mapDentistTypedInHolderElement(healthProvider)));

        return dentistTypedInHolderList;
    }

    private static DentistTypedInHolder mapDentistTypedInHolderElement(HealthProvider healthProvider) {
        DentistTypedInHolder dentistTypedInHolder = new DentistTypedInHolder(new IdHolder(healthProvider.getId().getRoot(), healthProvider.getId().getExtension(), healthProvider.getId().getAssigningAuthorityName()));

        // According to "Afklaringer vedr. FSK CDA-profil 2.0" (https://www.nspop.dk/x/jhUaBg)
        // Use organisation name as given and empty as family

        // givens
        dentistTypedInHolder.addGiven(
                getOrganizationName(healthProvider.getOrganization()));

        // family
        dentistTypedInHolder.setFamily(""); // empty, but not null

        AD addr = healthProvider
                .getOrganization() // jaxb required
                .getAddrs() // jaxb list
                .stream()
                .reduce((jaxbElement, jaxbElement2) -> {
                    throw new IllegalStateException("Multiple elements found for Health Provider address");
                })
                .orElseThrow(() -> new IllegalStateException("Expected address for Health Provider"));

        String use = addr.getUses().isEmpty() ? null : addr.getUses().get(0);
        List<AdxpStreetAddressLine> streetAddressLines = getJAXBElementsFromName(AdxpStreetAddressLine.class, "urn:hl7-org:v3", "streetAddressLine", addr.getContent());
        AdxpPostalCode postalCode = getSingleJAXBElementFromName(AdxpPostalCode.class, "urn:hl7-org:v3", "postalCode", addr.getContent());
        AdxpCity city = getSingleJAXBElementFromName(AdxpCity.class, "urn:hl7-org:v3", "city", addr.getContent());
        AdxpCountry country = getSingleJAXBElementFromName(AdxpCountry.class, "urn:hl7-org:v3", "country", addr.getContent());

        // address
        dentistTypedInHolder.setAddress(
                mapAddressHolderSimple(
                        use,
                        streetAddressLines.stream()
                                .map(PDCDataMapper::binToString)
                                .collect(Collectors.toList()),
                        binToString(city),
                        binToString(postalCode),
                        binToString(country)));

        POCDMT000040Author dataEnterer = healthProvider
                .getDataEnterer(); // jaxb required

        POCDMT000040AssignedAuthor assignedAuthor = dataEnterer
                .getAssignedAuthor(); // jaxb required

        // time
        dentistTypedInHolder.setAuthorTime(getDataEntererTime(dataEnterer));

        // authorCpr
        dentistTypedInHolder.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor
                .getAssignedPerson(); // jaxb optional

        // authorGivens
        getPersonGivens(assignedPerson)
                .forEach(dentistTypedInHolder::addAuthorGiven);

        // Optional yder
        healthProvider.getOrganization()
                .getIds()
                .stream()
                .filter(x -> x.getAssigningAuthorityName().equals("Yder"))
                .findFirst()
                .ifPresent(x -> dentistTypedInHolder.setYder(x.getExtension()));

        // Optional sor
        healthProvider.getOrganization()
                .getIds()
                .stream()
                .filter(x -> x.getAssigningAuthorityName().equals("SOR"))
                .findFirst()
                .ifPresent(x -> dentistTypedInHolder.setSor(x.getExtension()));

        // authorFamily
        dentistTypedInHolder.setAuthorFamily(getPersonFamily(assignedPerson));

        return dentistTypedInHolder;
    }

    private static List<TempAddrTypedInHolder> mapTempAddrTypedInHolder(RegisterData registerData) {
        List<TempAddrTypedInHolder> tempAddrTypedInHolderList = new ArrayList<>();

        getSKREntryElementsOfType(TemporaryAddress.class, registerData.getPersonalDataCard())
                .forEach(temporaryAddress ->
                        tempAddrTypedInHolderList.add(mapTempAddrTypedInHolderElement(temporaryAddress)));

        return tempAddrTypedInHolderList;
    }

    private static TempAddrTypedInHolder mapTempAddrTypedInHolderElement(TemporaryAddress temporaryAddress) {
        TempAddrTypedInHolder tempAddrTypedInHolder = new TempAddrTypedInHolder(new IdHolder(temporaryAddress.getId().getRoot(), temporaryAddress.getId().getExtension(), temporaryAddress.getId().getAssigningAuthorityName()));

        AD addr = temporaryAddress.getAddr();

        List<SXCMTS> usablePeriodElements = getJAXBElementsFromName(SXCMTS.class, "urn:hl7-org:v3", "useablePeriod", addr.getContent());

        if (usablePeriodElements.size() >= 1) {
            // timeFrom
            Optional.ofNullable(usablePeriodElements.get(0).getValue())
                    .ifPresent(value ->
                            tempAddrTypedInHolder.setPeriodFrom(
                                    LocalDate.parse(value, dateFormatter))); // expects date format (not date time)
        }

        if (usablePeriodElements.size() == 2) {
            // timeFrom
            Optional.ofNullable(usablePeriodElements.get(1).getValue())
                    .ifPresent(value ->
                            tempAddrTypedInHolder.setPeriodTo(
                                    LocalDate.parse(value, dateFormatter))); // expects date format (not date time)
        }

        List<String> uses = addr.getUses();
        String use = uses.isEmpty() ? null : uses.get(0);
        List<AdxpStreetAddressLine> streetAddressLines = getJAXBElementsFromName(AdxpStreetAddressLine.class, "urn:hl7-org:v3", "streetAddressLine", addr.getContent());
        AdxpPostalCode postalCode = getSingleJAXBElementFromName(AdxpPostalCode.class, "urn:hl7-org:v3", "postalCode", addr.getContent());
        AdxpCity city = getSingleJAXBElementFromName(AdxpCity.class, "urn:hl7-org:v3", "city", addr.getContent());
        AdxpCountry country = getSingleJAXBElementFromName(AdxpCountry.class, "urn:hl7-org:v3", "country", addr.getContent());

        // address
        tempAddrTypedInHolder.setAddress(
                mapAddressHolderSimple(
                        use,
                        streetAddressLines.stream()
                                .map(PDCDataMapper::binToString)
                                .collect(Collectors.toList()),
                        binToString(city),
                        binToString(postalCode),
                        binToString(country)));

        POCDMT000040Author dataEnterer = temporaryAddress
                .getDataEnterer(); // jaxb required

        POCDMT000040AssignedAuthor assignedAuthor = dataEnterer
                .getAssignedAuthor(); // jaxb required

        // time
        tempAddrTypedInHolder.setAuthorTime(getDataEntererTime(dataEnterer));

        // authorCpr
        tempAddrTypedInHolder.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor
                .getAssignedPerson(); // jaxb optional

        // authorGivens
        getPersonGivens(assignedPerson)
                .forEach(tempAddrTypedInHolder::addAuthorGiven);

        // authorFamily
        tempAddrTypedInHolder.setAuthorFamily(getPersonFamily(assignedPerson));

        return tempAddrTypedInHolder;
    }

    private static String binToString(BIN bin) {
        return Optional.ofNullable(bin)
                .flatMap(value ->
                        value.getContent()
                                .stream()
                                .reduce((serializable, serializable2) -> {
                                    throw new IllegalStateException("Multiple elements for string content");
                                })
                                .map(Objects::toString)
                )
                .orElse(null);
    }

    private static List<LanguageTypedInHolder> mapLanguageTypedInHolder(RegisterData registerData) {
        List<LanguageTypedInHolder> languageTypedInHolderList = new ArrayList<>();

        getSKREntryElementsOfType(Language.class, registerData.getPersonalDataCard())
                .forEach(language ->
                        languageTypedInHolderList.add(mapLanguageTypedInHolderElement(language)));

        return languageTypedInHolderList;
    }

    private static LanguageTypedInHolder mapLanguageTypedInHolderElement(Language language) {
        LanguageTypedInHolder languageTypedInHolder = new LanguageTypedInHolder(new IdHolder(language.getId().getRoot(), language.getId().getExtension(), language.getId().getAssigningAuthorityName()));

        // language code
        String languageCode = language.getLanguageCode() // jaxb required
                .getValue();
        languageTypedInHolder.setLanguageCode(
                languageCode);

        // language name
        languageTypedInHolder.setLanguageName(
                new Locale(languageCode).getDisplayName());

        POCDMT000040Author dataEnterer = language
                .getDataEnterer(); // jaxb required

        POCDMT000040AssignedAuthor assignedAuthor = dataEnterer
                .getAssignedAuthor(); // jaxb required

        // time
        languageTypedInHolder.setAuthorTime(getDataEntererTime(dataEnterer));

        // author cpr
        languageTypedInHolder.setAuthorCpr(getAssignedAuthorCpr(assignedAuthor));

        POCDMT000040Person assignedPerson = assignedAuthor
                .getAssignedPerson(); // jaxb optional

        // author givens
        getPersonGivens(assignedPerson)
                .forEach(languageTypedInHolder::addAuthorGiven);

        // author family
        languageTypedInHolder.setAuthorFamily(getPersonFamily(assignedPerson));

        return languageTypedInHolder;
    }

    private static <T> List<T> getJAXBElementsFromName(Class<T> clazz, String ns, String localPart, List<Serializable> elements) {
        return elements.stream()
                .filter(o -> o instanceof JAXBElement)
                .map(JAXBElement.class::cast)
                .filter(jaxbElement ->
                        jaxbElement.getName().getNamespaceURI().equals(ns)
                                && jaxbElement.getName().getLocalPart().equals(localPart))
                .map(JAXBElement::getValue)
                .filter(clazz::isInstance)
                .map(clazz::cast)
                .collect(Collectors.toList());
    }

    private static <T> T getSingleJAXBElementFromName(Class<T> clazz, String ns, String localPart, List<Serializable> elements) {
        return elements.stream()
                .filter(o -> o instanceof JAXBElement)
                .map(JAXBElement.class::cast)
                .filter(jaxbElement ->
                        jaxbElement.getName().getNamespaceURI().equals(ns)
                                && jaxbElement.getName().getLocalPart().equals(localPart))
                .reduce((jaxbElement, jaxbElement2) -> {
                    throw new IllegalStateException("Multiple element found with the name " + localPart);
                })
                .map(JAXBElement::getValue)
                .filter(clazz::isInstance)
                .map(clazz::cast)
                .orElse(null);
    }

    private static String getPersonFamily(POCDMT000040Person person) {
        return person.getNames()
                .stream()
                .reduce((pn, pn2) -> {
                    throw new IllegalStateException("Multiple name elements for person");
                })
                .flatMap(elements -> elements.getContent()
                        .stream()
                        .filter(o -> o instanceof JAXBElement)
                        .map(o -> ((JAXBElement<?>) o).getValue())
                        .filter(o -> o instanceof EnFamily)
                        .map(o -> ((EnFamily) o).getContent())
                        .flatMap(Collection::stream)
                        .map(Objects::toString)
                        .reduce((pn, pn2) -> {
                            throw new IllegalStateException("Multiple name->family elements for person");
                        }))
                .orElse(null);
    }

    public static List<String> getPersonGivens(POCDMT000040Person person) {
        return person.getNames()
                .stream()
                .reduce((pn, pn2) -> {
                    throw new IllegalStateException("Multiple name elements for person");
                })
                .map(pn -> pn.getContent()
                        .stream()
                        .filter(o -> o instanceof JAXBElement)
                        .map(o -> ((JAXBElement<?>) o).getValue())
                        .filter(o -> o instanceof EnGiven)
                        .map(o -> ((EnGiven) o).getContent())
                        .flatMap(Collection::stream)
                        .map(Objects::toString)
                        .collect(Collectors.toList()))
                .orElse(Collections.emptyList());
    }

    public static String getOrganizationName(POCDMT000040Organization organization) {
        // jaxb list
        return organization.getNames() // jaxb list
                .stream()
                .reduce((pn, pn2) -> {
                    throw new IllegalStateException("Multiple name elements for organization");
                })
                .flatMap(on -> on.getContent()
                        .stream()
                        .reduce((serializable, serializable2) -> {
                            throw new IllegalStateException("Multiple elements for string content");
                        })
                        .map(Objects::toString))
                .orElse(null);
    }

    private static PhoneHolder mapPhoneHolder(TEL tel) {
        PhoneHolder phoneHolder = new PhoneHolder();

        // uses
        tel.getUses()
                .forEach(phoneHolder::addUse);

        // value
        phoneHolder.setValue(tel.getValue());

        return phoneHolder;
    }

    public static String getAssignedAuthorCpr(POCDMT000040AssignedAuthor assignedAuthor) {
        return getIdElementByRoot("1.2.208.176.1.2", assignedAuthor)
                .map(II::getExtension)
                .orElse(null);
    }

    public static ZonedDateTime getDataEntererTime(POCDMT000040Author dataEnterer) {
        String timeValue = dataEnterer
                .getTime() // jaxb required
                .getValue();
        return Optional.ofNullable(timeValue)
                .map(value ->
                        ZonedDateTime.parse(
                                value, dateTimeFormatter))
                .orElse(null);
    }

    public static Optional<II> getIdElementByRoot(String root, POCDMT000040AssignedAuthor assignedAuthor) {
        return assignedAuthor.getIds() // schema requires 1 or more
                .stream()
                .filter(ii -> ii.getRoot().equals(root))
                .reduce((ignored, ignored2) -> {
                    throw new IllegalStateException("Multiple id elements with same root");
                });
    }

    private static <T> List<T> getSKREntryElementsOfType(Class<T> clazz, GetPersonalDataCardResponse getPersonalDataCardResponse) {

        if (Optional.ofNullable(getPersonalDataCardResponse)
                .map(GetPersonalDataCardResponse::getClinicalDocument)
                .map(POCDMT000040ClinicalDocument::getComponent)
                .map(POCDMT000040Component2::getStructuredBody)
                .map(POCDMT000040StructuredBody::getComponents).isPresent()) {

            POCDMT000040Component3 component3 = Optional.ofNullable(getPersonalDataCardResponse)
                    .map(GetPersonalDataCardResponse::getClinicalDocument)
                    .map(POCDMT000040ClinicalDocument::getComponent)
                    .map(POCDMT000040Component2::getStructuredBody)
                    .map(POCDMT000040StructuredBody::getComponents)
                    .get()
                    .get(0);

            if (component3 != null) {
                List<T> returnEntries = new ArrayList<T>();
                List<POCDMT000040Entry> entries = component3.getSection().getEntries();

                for (int i = 0; i < entries.size(); i++) {
                    POCDMT000040Entry entry = entries.get(i);
                    if (clazz.equals(Language.class)) {
                        if (entry.getLanguage() != null) {
                            returnEntries.add((T) entry.getLanguage());
                        }
                    } else if (clazz.equals(TemporaryAddress.class)) {
                        if (entry.getTemporaryAddress() != null) {
                            returnEntries.add((T) entry.getTemporaryAddress());
                        }
                    } else if (clazz.equals(HealthProvider.class)) {
                        if (entry.getHealthProvider() != null) {
                            returnEntries.add((T) entry.getHealthProvider());
                        }
                    } else if (clazz.equals(PatientContact.class)) {
                        if (entry.getPatientContact() != null) {
                            returnEntries.add((T) entry.getPatientContact());
                        }
                    } else if (clazz.equals(RelatedPerson.class)) {
                        if (entry.getRelatedPerson() != null) {
                            returnEntries.add((T) entry.getRelatedPerson());
                        }
                    } else {
                        return Collections.emptyList();
                    }
                }

                return returnEntries;
            }

        }
        return Collections.emptyList();
    }

    private static LivingWillRegistrationHolder mapLivingWillRegistrationHolder(RegisterData registerData) {
        LivingWillRegistrationHolder livingWillRegistrationHolder = new LivingWillRegistrationHolder();

        // hasRegistration
        livingWillRegistrationHolder.setHasRegistration(registerData.getHasLivingWill());

        return livingWillRegistrationHolder;
    }

    private static TreatmentWillRegistrationHolder mapTreatmentWillRegistrationHolder(RegisterData registerData) {
        TreatmentWillRegistrationHolder treatmentWillRegistrationHolder = new TreatmentWillRegistrationHolder();

        // hasRegistration
        treatmentWillRegistrationHolder.setHasRegistration(registerData.getHasTreatmentWill());

        return treatmentWillRegistrationHolder;
    }

    private static OrganDonorRegistrationHolder mapOrganDonorRegistrationHolder(RegisterData registerData) {
        OrganDonorRegistrationHolder organDonorRegistrationHolder = new OrganDonorRegistrationHolder();

        // hasRegistration
        organDonorRegistrationHolder.setHasRegistration(registerData.getHasOrganDonorRegistration());

        return organDonorRegistrationHolder;
    }

    private static ProviderOrganizationHolder mapProviderOrganizationHolder(YderVO yder) {
        if (yder == null) {
            return null;
        }

        ProviderOrganizationHolder providerOrganizationHolder = new ProviderOrganizationHolder();

        // yderId
        providerOrganizationHolder.setYderId(yder.getYdernr());

        // sor
//        providerOrganizationHolder.setSor();

        // name
        providerOrganizationHolder.setName(yder.getName());

        // phones
        providerOrganizationHolder.addPhone(
                mapPhoneHolder(yder.getPhone()));

        // emails
        providerOrganizationHolder.addEmail(yder.getEmail());

        // address
        providerOrganizationHolder.setAddress(mapAddressHolderSimple("H", Collections.singletonList(yder.getAddress()), yder.getCity(), yder.getPostcode(), null));

        return providerOrganizationHolder;
    }

    @SuppressWarnings("SameParameterValue")
    private static AddressHolder mapAddressHolderSimple(String use, List<String> addressLines, String city, String postcode, String country) {
        AddressHolder addressHolder = new AddressHolder();

        // use
        addressHolder.setUse(use);

        // address line
        addressLines.forEach(addressHolder::addAddressLine);

        // postal code
        addressHolder.setPostalCode(postcode);

        // city
        addressHolder.setCity(city);

        addressHolder.setCountry(country);

        return addressHolder;
    }

    private static PhoneHolder mapPhoneHolder(String phone) {
        PhoneHolder phoneHolder = new PhoneHolder();

        // use
        phoneHolder.addUse("WP");

        // value
        phoneHolder.setValue("tel:" + phone);

        return phoneHolder;
    }

    public static PrimaryPersonInformationHolder mapPersonHolder(RegisterData registerData) {
        String cpr = registerData.getCprnr();

        PrimaryPersonInformationHolder personHolder = new PrimaryPersonInformationHolder();

        // cpr
        personHolder.setCpr(cpr);

        RegularCPRPersonType regularCPRPerson = registerData.getScesData()
                .getPersonWithHealthCareInformationStructure() // jaxb required
                .getPersonInformationStructure() // jaxb required
                .getRegularCPRPerson(); // jaxb required

        PersonNameStructureType personNameStructure = regularCPRPerson
                .getSimpleCPRPerson() // jaxb required
                .getPersonNameStructure(); // jaxb optional

        if (personNameStructure != null) {
            // givens
            // Split given names by whitespace
            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonGivenName()))) // jaxb required
                    .forEach(personHolder::addGiven);

            Arrays.asList(
                    ArrayUtils.nullToEmpty(
                            StringUtils.split(
                                    personNameStructure.getPersonMiddleName()))) // jaxb optional
                    .forEach(personHolder::addGiven); // add middle name to given names

            // family
            personHolder.setFamily(
                    personNameStructure.getPersonSurnameName()); // jaxb required
        }

        if (cpr != null && cpr.length() == 10) {
            boolean male = "13579".contains(cpr.substring(9)); // is 10th char 1, 3, 5, 7 or 9?

            // gender
            personHolder.setGender(male ? "M" : "F");
        }

        XMLGregorianCalendar birthDate = regularCPRPerson
                .getPersonBirthDateStructure() // jaxb required
                .getBirthDate(); // jaxb required

        // birthTime
        personHolder.setBirthDate(birthDate.toGregorianCalendar().toZonedDateTime().toLocalDate());

        // address
        personHolder.setAddress(mapAddressHolder(registerData));

        return personHolder;
    }

    public static AddressHolder mapAddressHolder(RegisterData registerData) {
        AddressHolder addressHolder = new AddressHolder();

        AddressCompleteType addressComplete = Optional.ofNullable(registerData)
                .map(RegisterData::getScesData) // jaxb required
                .map(GetPersonWithHealthCareInformationOut::getPersonWithHealthCareInformationStructure) // jaxb required
                .map(PersonWithHealthCareInformationStructureType::getPersonInformationStructure) // jaxb required
                .map(PersonInformationStructureType::getPersonAddressStructure) // jaxb required
                .get().getAddressComplete(); // jaxb optional


        if (addressComplete != null) {
            AddressPostalType addressPostal = addressComplete.getAddressPostal(); // jaxb required

            // use
            if (Optional.ofNullable(registerData)
                    .map(RegisterData::getScesData)
                    .map(GetPersonWithHealthCareInformationOut::getPersonWithHealthCareInformationStructure)
                    .map(PersonWithHealthCareInformationStructureType::getPersonInformationStructure)
                    .map(PersonInformationStructureType::getRegularCPRPerson)
                    .map(RegularCPRPersonType::isPersonInformationProtectionIndicator).get()) {
                addressHolder.setUse("CONF");
            } else {
                addressHolder.setUse("H");
            }

            if (addressPostal != null) {
                // address line 1
                addressHolder.addAddressLine(
                        buildAddressLine1(
                                addressPostal.getStreetName(), // jaxb required
                                addressPostal.getStreetBuildingIdentifier(), // jaxb optional
                                addressPostal.getFloorIdentifier(), // jaxb optional
                                addressPostal.getSuiteIdentifier())); // jaxb optional

                // address line 2
                Optional.ofNullable(
                        addressPostal.getPostOfficeBoxIdentifier()) // jaxb optional
                        .ifPresent(integer ->
                                addressHolder.addAddressLine(
                                        String.valueOf(integer)));

                // address line 3
                Optional.ofNullable(addressPostal.getDistrictSubdivisionIdentifier()) // jaxb optional
                        .ifPresent(addressHolder::addAddressLine);

                // address line 4
                Optional.ofNullable(addressPostal.getMailDeliverySublocationIdentifier()) // jaxb optional
                        .ifPresent(addressHolder::addAddressLine);

                // postal code
                addressHolder.setPostalCode(
                        addressPostal.getPostCodeIdentifier()); // jaxb required

                // city
                addressHolder.setCity(
                        addressPostal.getDistrictName()); // jaxb required

                // country
                Optional.ofNullable(addressPostal)
                        .map(AddressPostalType::getCountryIdentificationCode)
                        .map(CountryIdentificationCodeType::getValue)
                        .ifPresent(addressHolder::setCountry);
            }
        }

        return addressHolder;
    }

    private static String buildAddressLine1(String streetName, String streetBuildingIdentifier, String floor, String suiteId) {
        return streetName + withSpace(streetBuildingIdentifier) + withSpace(floor) + withSpace(suiteId);
    }

    private static String withSpace(String value) {
        if (value == null) {
            return "";
        } else {
            return " " + value;
        }
    }
}
