/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda.assembler;

import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.stamkort.AssigningAuthorities;
import dk.stamkort.OID;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.sundhedsdatastyrelsen.fsk._2019._08._01.GetPersonalDataCardResponse;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import oio.medcom.cprservice._1_0.*;
import org.apache.commons.lang.StringUtils;
import org.hl7.v3.ObjectFactory;
import org.hl7.v3.*;

import java.math.BigInteger;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static dk.stamkort.OID.STAMKORT_CODE_SYSTEM;

public class PDC_1_0ClinicalDocumentAssembler {
    private static final String HOME_ADDRESS_USE = "H";
    private static final String CONFIDENTIAL_ADDRESS_USE = "CONF";

    private static final String TS_FORMAT = "yyyyMMddHHmmssZ"; // Output sample: 20171019144200+0200

    private static final String emailRegEx = "(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|\"(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21\\x23-\\x5b\\x5d-\\x7f]|\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])*\")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21-\\x5a\\x53-\\x7f]|\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])+)\\])";
    private static final Pattern emailPattern = java.util.regex.Pattern.compile(emailRegEx);

    private static ObjectFactory o = new ObjectFactory();

    public static POCDMT000040ClinicalDocument assembleClinicalDocument(RegisterData registerData, String yderNr) {
        POCDMT000040ClinicalDocument doc = createDocument();

        insertPatientData(doc, registerData.getCprnr(), registerData.getScesData(), registerData.getYder(yderNr));
        insertPersonalDataCard(doc, registerData);
        insertOrganDonorRegistration(doc, registerData.getHasOrganDonorRegistration());
        insertLivingWill(doc, registerData.getHasLivingWill());
        insertTreatmentWill(doc, registerData.getHasTreatmentWill());

        if (doc.getComponent() != null && doc.getComponent().getStructuredBody() == null) {
            addEmptyComponent(doc);
        }

        return doc;
    }

    private static void insertLivingWill(POCDMT000040ClinicalDocument doc, Boolean hasLivingWill) {
        if (hasLivingWill != null) {
            POCDMT000040Component3 component = ensureSection(doc);
            addEntry(component, hasLivingWill, "LivingWill", STAMKORT_CODE_SYSTEM, "Living Will");
        }
    }

    private static void insertTreatmentWill(POCDMT000040ClinicalDocument doc, Boolean hasTreatmentWill) {
        if (hasTreatmentWill != null) {
            POCDMT000040Component3 component = ensureSection(doc);
            addEntry(component, hasTreatmentWill, "TreatmentWill", STAMKORT_CODE_SYSTEM, "Treatment Will");
        }
    }

    private static void insertOrganDonorRegistration(POCDMT000040ClinicalDocument doc, Boolean hasOrganDonorRegistration) {
        if (hasOrganDonorRegistration != null) {
            POCDMT000040Component3 component = ensureSection(doc);
            addEntry(component, hasOrganDonorRegistration, "OrganDonor", STAMKORT_CODE_SYSTEM, "OrganDonorRegistration");
        }
    }

    private static void addEntry(POCDMT000040Component3 component, Boolean assignedValue, String codeStr, String codeSystem, String displayName) {
        POCDMT000040Entry entry = new POCDMT000040Entry();
        POCDMT000040Observation observation = new POCDMT000040Observation();
        observation.setMoodCode(XActMoodDocumentObservation.EVN);
        observation.getClassCodes().add("OBS");
        CD code = new CD();
        code.setCode(codeStr);
        code.setDisplayName(displayName);
        code.setCodeSystem(codeSystem);
        observation.setCode(code);
        BN value = new BN();
        value.setValue(assignedValue);
        observation.getValues().add(value);
        entry.setObservation(observation);
        component.getSection().getEntries().add(entry);
    }

    private static POCDMT000040Component3 ensureSection(POCDMT000040ClinicalDocument doc) {
        POCDMT000040Component2 cdComponent = doc.getComponent();
        if (cdComponent == null) {
            cdComponent = new POCDMT000040Component2();
            doc.setComponent(cdComponent);
        }
        POCDMT000040StructuredBody structuredBody = doc.getComponent().getStructuredBody();
        if (structuredBody == null) {
            structuredBody = new POCDMT000040StructuredBody();
            doc.getComponent().setStructuredBody(structuredBody);
        }
        List<POCDMT000040Component3> components = structuredBody.getComponents();
        if (components.size() == 0) {
            POCDMT000040Component3 component = new POCDMT000040Component3();
            components.add(component);
        }
        POCDMT000040Component3 component = components.get(0);
        if (component.getSection() == null) {
            POCDMT000040Section section = new POCDMT000040Section();
            component.setSection(section);
        }
        return component;
    }

    private static POCDMT000040ClinicalDocument createDocument() {
        POCDMT000040ClinicalDocument doc = o.createPOCDMT000040ClinicalDocument();

        doc.setId(createId(OID.ISO_MEMBERBODY_DK_MEDCOM, "NA", "MedCom"));
        II templateId = new II();
        templateId.setRoot("1.2.208.184.15.1");
        doc.getTemplateIds().add(templateId);
        doc.setCode(createCode("NA", OID.STAMKORT_CODE_SYSTEM, "Stamkort"));
        doc.setTypeId(createTypeId(OID.HL7_REGISTEREDE_RMIMS, "POCD_HD000040"));
        doc.setConfidentialityCode(createCode("N", OID.CONFIDENTIALITYCODE));
        doc.setLanguageCode(createCode("da-DK"));
        doc.setTitle(createText("Det Fælles StamKort"));
        doc.setClassCode(ActClinicalDocument.DOCCLIN);
        doc.getMoodCodes().add("EVN");
        CS dk = new CS();
        dk.setCode("DK");
        doc.getRealmCodes().add(dk);

        doc.setEffectiveTime(createTimestamp(Instant.now()));
        doc.setVersionNumber(createVersionNumber(0));
        doc.setCustodian(createCustodian());
        doc.getAuthors().add(createEmptyAuthor());
        //addEmptyComponent(doc);

        return doc;
    }

    private static INT createVersionNumber(int version) {
        INT versionNumber = o.createINT();
        versionNumber.setValue(BigInteger.valueOf(version));
        return versionNumber;
    }

    public static II createId(String root, String extension, String assigningAuth) {
        II id = o.createII();
        if (root != null) {
            id.setRoot(root);
        }
        if (extension != null) {
            id.setExtension(extension);
        }
        if (assigningAuth != null) {
            id.setAssigningAuthorityName(assigningAuth);
        }
        return id;
    }

    private static CS createCode(String code) {
        CS c = o.createCS();
        c.setCode(code);
        return c;
    }

    private static CE createCode(String code, String codeSystem) {
        return createCode(code, codeSystem, null);
    }

    private static CE createCode(String code, String codeSystem, String displayName) {
        CE c = o.createCE();
        c.setCode(code);
        c.setCodeSystem(codeSystem);
        c.setDisplayName(displayName);
        return c;
    }

    private static POCDMT000040InfrastructureRootTypeId createTypeId(String root, String extension) {
        POCDMT000040InfrastructureRootTypeId typeId = new POCDMT000040InfrastructureRootTypeId();
        typeId.setRoot(root);
        typeId.setExtension(extension);
        return typeId;
    }

    private static ST createText(String text) {
        ST st = o.createST();
        st.getContent().add(text);
        return st;
    }

    private static TS createTimestamp(Instant instant) {
        TS ts = o.createTS();
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(TS_FORMAT).withZone(ZoneId.systemDefault());
        String format = formatter.format(Instant.now());
        ts.setValue(format);
        return ts;
    }

    private static POCDMT000040Custodian createCustodian() {
        POCDMT000040Custodian custodian = o.createPOCDMT000040Custodian();
        POCDMT000040AssignedCustodian assignedCustodian = o.createPOCDMT000040AssignedCustodian();
        POCDMT000040CustodianOrganization custodianOrganisation = o.createPOCDMT000040CustodianOrganization();
        II id = o.createII();
        id.setRoot("NA");
        custodianOrganisation.getIds().add(id);
        assignedCustodian.setRepresentedCustodianOrganization(custodianOrganisation);
        custodian.setAssignedCustodian(assignedCustodian);

        return custodian;
    }

    private static POCDMT000040Author createEmptyAuthor() {
        POCDMT000040Author author = o.createPOCDMT000040Author();

        POCDMT000040AssignedAuthor assignedAuthor = o.createPOCDMT000040AssignedAuthor();
        assignedAuthor.getIds().add(createId("NA", null, null));
        POCDMT000040Person assignedPerson = new POCDMT000040Person();
        assignedPerson.getNullFlavors().add("NA");
        assignedAuthor.setAssignedPerson(assignedPerson);
        author.setAssignedAuthor(assignedAuthor);
        TS time = new TS();
        time.getNullFlavors().add("NA");
        author.setTime(time);

        return author;
    }

    // Component is required. We just add an empty "nonXMLBody"
    private static void addEmptyComponent(POCDMT000040ClinicalDocument clinicalDocument) {
        POCDMT000040Component2 component = new POCDMT000040Component2();
        POCDMT000040NonXMLBody nonXMLBody = new POCDMT000040NonXMLBody();
        ST st = new ST();
        st.getNullFlavors().add("NA");
        nonXMLBody.setText(st);
        component.setNonXMLBody(nonXMLBody);
        clinicalDocument.setComponent(component);
    }

    private static POCDMT000040Patient createPatient(String cprnr, String givenName, String familyName) {
        POCDMT000040Patient patient = o.createPOCDMT000040Patient();
        patient.getClassCodes().add("PSN");
        patient.setDeterminerCode("INSTANCE");

        patient.getNames().add(createPersonName(givenName, familyName));

        if (cprnr != null && cprnr.length() == 10) {
            boolean male = "13579".contains(cprnr.substring(9));
            CE genderCode = createCode(male ? "M" : "F", OID.ADMINISTRATIVE_GENDER);
            patient.setAdministrativeGenderCode(genderCode);
        }

        return patient;
    }

    public static PN createPersonName(String givenName, String familyName) {
        PN name = o.createPN();
        EnGiven given = o.createEnGiven();
        given.getContent().add(givenName);
        name.getContent().add(o.createENGiven(given));
        EnFamily family = o.createEnFamily();
        family.getContent().add(familyName);
        name.getContent().add(o.createENFamily(family));

        return name;
    }

    private static AD mapAddress(GetPersonWithHealthCareInformationOut person, RegularCPRPersonType regularCPRPersonType) {
        AddressPostalType addressPostal = person.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal();
        String streetName = addressPostal.getStreetName();
        String streetBuildingIdentifier = addressPostal.getStreetBuildingIdentifier();
        String floor = addressPostal.getFloorIdentifier();
        String suiteId = addressPostal.getSuiteIdentifier();
        String postCode = addressPostal.getPostCodeIdentifier();
        String districtName = addressPostal.getDistrictName();
        String districtSubdivision = addressPostal.getDistrictSubdivisionIdentifier();
        String mailDeliverySublocation = addressPostal.getMailDeliverySublocationIdentifier();
        String countryIdentificationCode = addressPostal.getCountryIdentificationCode() == null ?
                null : addressPostal.getCountryIdentificationCode().getValue();
        Integer postOfficeBoxIdentifier = addressPostal.getPostOfficeBoxIdentifier();
        String addressLine1 = buildAddressLine1(streetName, streetBuildingIdentifier, floor, suiteId);
        String addressLine2 = postOfficeBoxIdentifier == null ? null : "" + postOfficeBoxIdentifier;
        return createAddress(addressLine1, addressLine2, districtSubdivision, mailDeliverySublocation, postCode, districtName, countryIdentificationCode, regularCPRPersonType);
    }

    private static String buildAddressLine1(String streetName, String streetBuildingIdentifier, String floor, String suiteId) {
        return streetName + withSpace(streetBuildingIdentifier) + withSpace(floor) + withSpace(suiteId);
    }

    private static String withSpace(String value) {
        if (value == null) {
            return "";
        } else {
            return " " + value;
        }
    }

    private static AD createAddress(String addressLine1, String addressLine2, String addressLine3, String addressLine4, String postalCode, String city, String country, RegularCPRPersonType regularCPRPersonType) {
        AD address = o.createAD();
        address.setIsNotOrdered(false);

        // Do not change the ordering of the street address line elements below. The parsing is dependent on the ordering.

        if (addressLine1 != null) {
            AdxpStreetAddressLine adxpStreetAddressLine = o.createAdxpStreetAddressLine();
            adxpStreetAddressLine.getContent().add(addressLine1);
            address.getContent().add(o.createADStreetAddressLine(adxpStreetAddressLine));
        }

        if (addressLine2 != null) {
            AdxpStreetAddressLine adxpStreetAddressLine = o.createAdxpStreetAddressLine();
            adxpStreetAddressLine.getContent().add(addressLine2);
            address.getContent().add(o.createADStreetAddressLine(adxpStreetAddressLine));
        }

        if (addressLine3 != null) {
            AdxpStreetAddressLine adxpStreetAddressLine = o.createAdxpStreetAddressLine();
            adxpStreetAddressLine.getContent().add(addressLine3);
            address.getContent().add(o.createADStreetAddressLine(adxpStreetAddressLine));
        }

        if (addressLine4 != null) {
            AdxpStreetAddressLine adxpStreetAddressLine = o.createAdxpStreetAddressLine();
            adxpStreetAddressLine.getContent().add(addressLine4);
            address.getContent().add(o.createADStreetAddressLine(adxpStreetAddressLine));
        }

        if (postalCode != null) {
            AdxpPostalCode adxpPostalCode = o.createAdxpPostalCode();
            adxpPostalCode.getContent().add(postalCode);
            address.getContent().add(o.createADPostalCode(adxpPostalCode));
        }

        if (city != null) {
            AdxpCity adxpCity = o.createAdxpCity();
            adxpCity.getContent().add(city);
            address.getContent().add(o.createADCity(adxpCity));
        }

        if (country != null) {
            AdxpCountry adxpCountry = o.createAdxpCountry();
            adxpCountry.getContent().add(country);
            address.getContent().add(o.createADCountry(adxpCountry));
        }

        if (regularCPRPersonType == null) {
            address.getUses().add(HOME_ADDRESS_USE);
            return address;
        }

        address.getUses().add(regularCPRPersonType.isPersonInformationProtectionIndicator() ? CONFIDENTIAL_ADDRESS_USE : HOME_ADDRESS_USE);
        return address;
    }

    /**
     * Insert name and address of person into document
     *
     * @param clinicalDocument Document to insert into
     * @param scesData         CPR data from SCES
     * @param yder
     */
    private static void insertPatientData(POCDMT000040ClinicalDocument clinicalDocument, String cprnr, GetPersonWithHealthCareInformationOut scesData, YderVO yder) {
        POCDMT000040RecordTarget recordTarget = o.createPOCDMT000040RecordTarget();
        recordTarget.getTypeCodes().add("RCT");
        recordTarget.setContextControlCode("OP");

        POCDMT000040PatientRole patientRole = o.createPOCDMT000040PatientRole();
        patientRole.getClassCodes().add("PAT");
        patientRole.getIds().add(createId(OID.CPR, cprnr, AssigningAuthorities.CPR));
        patientRole.getNullFlavors().add("NA");
        recordTarget.setPatientRole(patientRole);

        if (scesData != null) {
            // name and gender
            RegularCPRPersonType regularCPRPersonType = scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson();
            PersonNameStructureType personNameStructure = regularCPRPersonType.getSimpleCPRPerson().getPersonNameStructure();
            String personGivenName = personNameStructure.getPersonGivenName();
            String personMiddleName = personNameStructure.getPersonMiddleName();
            String personSurnameName = personNameStructure.getPersonSurnameName();
            patientRole.setPatient(createPatient(cprnr, personGivenName + (StringUtils.isEmpty(personMiddleName) ? "" : " " + personMiddleName), personSurnameName));
            recordTarget.setPatientRole(patientRole);

            // Sikringsgruppe
            PersonHealthCareInformationStructureType healthCareInformation = scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure();
            if (healthCareInformation != null) {
                PersonPublicHealthInsuranceType personPublicHealthInsurance = healthCareInformation.getPersonPublicHealthInsurance();
                if (personPublicHealthInsurance != null) {
                    if (personPublicHealthInsurance.getPublicHealthInsuranceGroupIdentifier() != null) {
                        recordTarget.getPatientRole().setTypeId(createTypeId(OID.SIKRINGSGRUPPE, personPublicHealthInsurance.getPublicHealthInsuranceGroupIdentifier().value()));
                    }
                }
            }
            // address
            AD address = mapAddress(scesData, regularCPRPersonType);
            patientRole.getAddrs().add(address);

            if (healthCareInformation != null) {
                AssociatedGeneralPractitionerStructureType associatedGeneralPractitioner = healthCareInformation.getCurrentGeneralPractitionerStructure().getAssociatedGeneralPractitionerStructure();
                if (associatedGeneralPractitioner != null) {
                    BigInteger ydernr = associatedGeneralPractitioner.getAssociatedGeneralPractitionerIdentifier();
                    if (ydernr != null && !ydernr.equals(new BigInteger("0")) && yder != null) {
                        // assign yder as a participant structure in the ClinicalDocument header
                        clinicalDocument.getParticipants().add(createOwnDoctor(yder));
                    }
                }
            }
        }

        clinicalDocument.getRecordTargets().add(recordTarget);
    }

    private static POCDMT000040Participant1 createOwnDoctor(YderVO yder) {
        POCDMT000040Participant1 participant = new POCDMT000040Participant1();
        participant.getTypeCodes().add("IND");
        II template = new II();
        template.setRoot(OID.PATIENT_CONTACT);
        participant.getTemplateIds().add(template);
        POCDMT000040AssociatedEntity associatedEntity = new POCDMT000040AssociatedEntity();
        associatedEntity.getClassCodes().add("CAREGIVER");
        CE code = new CE();
        code.setCode("YDERNUMMER");
        code.setCodeSystem(OID.YDERNUMMER);
        associatedEntity.setCode(code);
        POCDMT000040Organization scopingOrganization = new POCDMT000040Organization();
        II ydernrId = new II();
        ydernrId.setExtension(yder.getYdernr());
        ydernrId.setAssigningAuthorityName("Yderregisteret"); // http://oid-info.com/get/1.2.208.176.1.4
        ydernrId.setRoot(OID.YDERNUMMER);
        scopingOrganization.getIds().add(ydernrId);
        if (!StringUtils.isEmpty(yder.getName())) {
            ON orgName = new ON();
            orgName.getContent().add(yder.getName());
            scopingOrganization.getNames().add(orgName);
        }
        addYderTelecom(yder, scopingOrganization);
        AD address = createAddress(yder.getAddress(), null, null, null, yder.getPostcode(), yder.getCity(), null, null);
        scopingOrganization.getAddrs().add(address);

        associatedEntity.setScopingOrganization(scopingOrganization);
        participant.setAssociatedEntity(associatedEntity);
        return participant;
    }

    private static void addYderTelecom(YderVO yder, POCDMT000040Organization organization) {
        if (yder.getPhone() != null || yder.getEmail() != null) {
            if (!StringUtils.isEmpty(yder.getPhone())) {
                TEL phone = o.createTEL();
                phone.setValue("tel:" + yder.getPhone());
                phone.getUses().add("WP");
                organization.getTelecoms().add(phone);
            }
            if (!StringUtils.isEmpty(yder.getEmail()) && isEmail(yder.getEmail())) {
                TEL email = o.createTEL();
                email.setValue("mailto:" + yder.getEmail());
                email.getUses().add("WP");
                organization.getTelecoms().add(email);
            }
        }
    }

    static boolean isEmail(String email) {
        if (org.springframework.util.StringUtils.isEmpty(email)) {
            return false;
        }
        Matcher emailMatcher = emailPattern.matcher(email);
        boolean matches = emailMatcher.matches();
        return matches;
    }

    /**
     * Pull the following from PersonalDataCard register response:
     * - author - the last author of PersonalDataCard will be author of clinicalDocument
     * - all entries in structured body
     *
     * @param clinicalDocument
     * @param rd
     */
    private static void insertPersonalDataCard(POCDMT000040ClinicalDocument clinicalDocument, RegisterData rd) {
        GetPersonalDataCardResponse skrResponse = rd.getPersonalDataCard();
        if (skrResponse != null) {
            POCDMT000040ClinicalDocument doc = skrResponse.getClinicalDocument();
            if (doc != null) {
                if (doc.getComponent() != null && doc.getComponent().getStructuredBody() != null && doc.getComponent().getStructuredBody().getComponents() != null) {
                    for (POCDMT000040Component3 component : doc.getComponent().getStructuredBody().getComponents()) {
                        if (component.getSection() != null && component.getSection().getEntries() != null) {
                            for (POCDMT000040Entry entry : component.getSection().getEntries()) {
                                insertEntry(clinicalDocument, entry);
                                String ydernr = getYdernr(entry);
                                if (ydernr != null) {
                                    setYderMasterData(entry, rd.getYder(ydernr));
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    private static void setYderMasterData(POCDMT000040Entry entry, YderVO yder) {
        if (yder == null) {
            return;
        }
        POCDMT000040Organization organization = entry.getHealthProvider().getOrganization();

        // Address
        clearList(organization.getAddrs());
        AD address = createAddress(yder.getAddress(), null, null, null, yder.getPostcode(), yder.getCity(), null, null);
        organization.getAddrs().add(address);

        // Name
        clearList(organization.getNames());
        if (yder.getName() != null) {
            ON name = new ON();
            name.getContent().add(yder.getName());
            organization.getNames().add(name);
        }

        // Phone/Email
        clearList(organization.getTelecoms());
        addYderTelecom(yder, organization);
    }

    private static void clearList(List<?> list) {
        if (list.size() > 0) {
            list.clear();
        }
    }

    private static String getYdernr(POCDMT000040Entry entry) {
        if (entry == null) {
            return null;
        }
        HealthProvider healthProvider = entry.getHealthProvider();
        if (healthProvider == null) {
            return null;
        }
        POCDMT000040Organization organization = healthProvider.getOrganization();
        if (organization == null) {
            return null;
        }
        List<II> ids = organization.getIds();
        if (ids == null || ids.size() == 0) {
            return null;
        }
        II id = ids.get(0);
        if (id == null || StringUtils.isEmpty(id.getExtension())) {
            return null;
        }
        return id.getExtension();
    }

    private static void insertEntry(POCDMT000040ClinicalDocument doc, POCDMT000040Entry entry) {
        if (doc != null) {
            POCDMT000040Component2 component = doc.getComponent();
            if (component == null) {
                component = o.createPOCDMT000040Component2();
                doc.setComponent(component);
            }
            POCDMT000040StructuredBody structuredBody = component.getStructuredBody();
            if (structuredBody == null) {
                structuredBody = o.createPOCDMT000040StructuredBody();
                component.setStructuredBody(structuredBody);
            }
            POCDMT000040Component3 innerComponent = structuredBody.getComponents().size() > 0 ? structuredBody.getComponents().get(0) : null;
            if (innerComponent == null) {
                innerComponent = o.createPOCDMT000040Component3();
                structuredBody.getComponents().add(innerComponent);
            }
            POCDMT000040Section section = innerComponent.getSection();
            if (section == null) {
                section = o.createPOCDMT000040Section();
                innerComponent.setSection(section);
            }
            section.getEntries().add(entry);
        }
    }
}
