/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sds.dgws_idws_proxy.security.SOSIUtil;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.OrganizationTypeVO;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.integrations.minlog.MinLogDataWrapper;
import dk.stamkort.integrations.minlog.MinLogException;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.webservice.hl7cda.ClinicalDocumentProvider;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType;
import org.apache.commons.io.Charsets;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import javax.annotation.PostConstruct;
import javax.xml.bind.JAXBException;

@Component
public class DocumentProvider {
    private static final Logger log = LogManager.getLogger(DocumentProvider.class);
    private static final String SERVICE_IDENTIFIER = "Det Fælles Stamkort";

    @Autowired
    public PersonalDataCardServiceImpl personalDataCardService;

    @Autowired
    public RegistryIndexDAO registryIndexDAO;

    @Autowired
    private MinLogDataWrapper minLogDataWrapper;

    @Autowired
    private Properties properties;

    @Autowired
    private ClinicalDocumentProvider clinicalDocumentProvider;

    public boolean cvrIsWhiteListed(String cvr) {
        if (StringUtils.isEmpty(cvr)) {
            log.warn("CVR number is not provided for level 3: " + cvr);
            return false;
        }
        String whitelistedLevel3CVRs = properties.getWhitelistedLevel3CVRs();
        boolean contains = whitelistedLevel3CVRs.contains(cvr);
        if (!contains) {
            log.warn("CVR number is not allowed for level 3: " + cvr);
        }
        return contains;
    }

    @PostConstruct
    public void init() {
        DocumentProviderWS.documentProvider = this;
        DocumentProviderWSHandler.setSosiUtil(new SOSIUtil(properties.isProduction(), properties.isIgnoreInvalidIdcardInTestMode()));
    }

    public RetrieveDocumentSetResponseType.DocumentResponse getDocument(RetrieveDocumentSetRequestType.DocumentRequest documentRequest) throws JAXBException, DAOException {

        RetrieveDocumentSetResponseType.DocumentResponse documentResponse = new RetrieveDocumentSetResponseType.DocumentResponse();
        documentResponse.setRepositoryUniqueId(documentRequest.getRepositoryUniqueId());
        documentResponse.setDocumentUniqueId(documentRequest.getDocumentUniqueId());
        documentResponse.setHomeCommunityId(documentRequest.getHomeCommunityId());
        documentResponse.setMimeType("text/xml");

        String documentUniqueId = documentRequest.getDocumentUniqueId();
        RegistryIndexVO registryIndex = registryIndexDAO.getRegistryIndexByDocumentId(documentUniqueId);
        RequestContextVO.get().addAuditLogParameter("documentId", documentUniqueId);
        if (registryIndex == null) {
            return null;
        } else {
            RequestContextVO.get().addAuditLogParameter("citizenCpr", registryIndex.getPersonIdentifier());
        }

        String xml = clinicalDocumentProvider.getClinicalDocument(registryIndex.getPersonIdentifier(), registryIndex.getTypeCode(), registryIndex.getFormatCode());
        documentResponse.setDocument(xml.getBytes(Charsets.UTF_8));

        logReadActivityToMinLog(RequestContextVO.get(), registryIndex.getPersonIdentifier());

        return documentResponse;
    }

	private void logReadActivityToMinLog(RequestContextVO context, String cprnr) {
        try {
            ProxyData proxyData = context.getProxyData();

            if (proxyData != null) {
                minLogDataWrapper.logReadActivity(context, properties.getMinLogReadActivityText(), cprnr, proxyData.getCareProviderID(), proxyData.getCareProviderName(),
                        OrganizationTypeVO.fromMedcomType(proxyData.getCareProviderType()), proxyData.getCprnr(), null,
                        SERVICE_IDENTIFIER);
            }
        } catch (MinLogException e) {
            log.error("Error writing to MinLog.", e);
        }
    }
}
