/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.stamkort.auditlog.AuditLogger;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.integrations.dds.XdsDocumentEntryConstants;
import dk.stamkort.util.FSKException;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.util.RequestContextVO;
import ihe.iti.xds_b._2007.DocumentRepositoryPortType;
import ihe.iti.xds_b._2007.ProvideAndRegisterDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType;
import oasis.names.tc.ebxml_regrep.xsd.rs._3.RegistryError;
import oasis.names.tc.ebxml_regrep.xsd.rs._3.RegistryErrorList;
import oasis.names.tc.ebxml_regrep.xsd.rs._3.RegistryResponseType;
import org.apache.commons.lang.NotImplementedException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.springframework.web.context.support.SpringBeanAutowiringSupport;

import javax.annotation.PostConstruct;
import javax.jws.HandlerChain;
import javax.jws.WebService;
import javax.xml.bind.JAXBException;
import javax.xml.ws.BindingType;
import javax.xml.ws.Holder;
import javax.xml.ws.soap.Addressing;

import static dk.stamkort.integrations.dds.XdsDocumentEntryConstants.XDS_STATUS_FAILURE;
import static dk.stamkort.integrations.dds.XdsDocumentEntryConstants.XDS_STATUS_SUCCESS;
import static javax.xml.ws.soap.SOAPBinding.SOAP12HTTP_MTOM_BINDING;

@WebService(endpointInterface = "ihe.iti.xds_b._2007.DocumentRepositoryPortType", targetNamespace = "urn:ihe:iti:xds-b:2007", serviceName = "services/fsk", portName = "DocumentRepository_Port_Soap12")
@Addressing(enabled = true, required = false)
@BindingType(value = SOAP12HTTP_MTOM_BINDING)
@HandlerChain(file = "/documentprovider-handler.xml")
public class DocumentProviderWS implements DocumentRepositoryPortType {

    public static final String AUTH_LEVEL_SYSTEM = "3";

    private static final Logger log = LogManager.getLogger(DocumentProviderWS.class);

    public static DocumentProvider documentProvider;

    private static AuditLogger auditLogger = new AuditLogger(DocumentProviderWS.class);

    @PostConstruct
    public void init() {
        SpringBeanAutowiringSupport.processInjectionBasedOnCurrentContext(this);
    }

    // Required by wsdl, but will not be implemented in FSK
    //@Override
    public RegistryResponseType documentRepositoryProvideAndRegisterDocumentSetB(ProvideAndRegisterDocumentSetRequestType provideAndRegisterDocumentSetRequestType) {
        throw new NotImplementedException();
    }

    @SuppressWarnings("ValidExternallyBoundObject")
    @Override
    public RetrieveDocumentSetResponseType documentRepositoryRetrieveDocumentSet(
            RetrieveDocumentSetRequestType body,
            Security securityHeader,
            Holder<Header> medcomHeader,
            HsuidHeader hsuidHeader) {

        long startTime = System.currentTimeMillis();

        RequestContextVO requestContext = RequestContextVO.get();
        if (requestContext.getProxyData() == null) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER); // ensures that DocumentProviderWSHandler has validated security
        }
        if (!okForLevel3(requestContext)) {
            FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER); // ensures that DocumentProviderWSHandler has validated security
        }

        try {
            RetrieveDocumentSetResponseType responseType = new RetrieveDocumentSetResponseType();

            RegistryResponseType regResponse = new RegistryResponseType();
            for (RetrieveDocumentSetRequestType.DocumentRequest documentRequest : body.getDocumentRequest()) {
                try {
                    RetrieveDocumentSetResponseType.DocumentResponse documentResponse = documentProvider.getDocument(documentRequest);
                    if (documentResponse != null) {
                        responseType.getDocumentResponse().add(documentResponse);
                    } else {
                        RegistryError error = new RegistryError();
                        error.setCodeContext("Document not found for id " + documentRequest.getDocumentUniqueId());
                        error.setErrorCode(XdsDocumentEntryConstants.XDS_ERROR_CODE_MISSING_DOCUMENT);
                        ensureRegistryError(regResponse);
                        regResponse.getRegistryErrorList().getRegistryError().add(error);
                    }
                } catch (JAXBException|DAOException|FSKException e) {
                    log.warn("Unrecoverable exception", e);
                    ensureRegistryError(regResponse);
                    regResponse.getRegistryErrorList().getRegistryError().add(createRegistryError(e));
                }
            }

            if (regResponse.getRegistryErrorList() != null && regResponse.getRegistryErrorList().getRegistryError().size() > 0) {
                regResponse.setStatus(XDS_STATUS_FAILURE);
            } else {
                regResponse.setStatus(XDS_STATUS_SUCCESS);
            }

            responseType.setRegistryResponse(regResponse);
            auditLogger.auditOk(startTime, RequestContextVO.get().getProxyData());

            return responseType;
        } catch (RuntimeException e) {
            auditLogger.auditError(startTime, RequestContextVO.get().getProxyData(), e.getMessage(), e);
            throw e;
        }
    }

    boolean okForLevel3(RequestContextVO requestContext) {
        ProxyData proxyData = requestContext.getProxyData();
        if (AUTH_LEVEL_SYSTEM.equals(proxyData.getAuthenticationLevel())) {
            // Client must be whitelisted. Check if so...
            if (proxyData == null) {
                log.warn("No proxy data");
                return false;
            }
            if (!"medcom:cvrnumber".equals(proxyData.getCareProviderType())) {
                log.warn("Unsupported CareProviderType for level 3: " + proxyData.getCareProviderType());
                return false;
            }
            return documentProvider.cvrIsWhiteListed(proxyData.getCareProviderID());
        } else {
            // Not level 3. Whitelisting check is not relevant
            return true;
        }
    }

    private void ensureRegistryError(RegistryResponseType regResponse) {
        if (regResponse.getRegistryErrorList() == null) {
            regResponse.setRegistryErrorList(new RegistryErrorList());
        }
    }

    // XDS error codes: http://www.medicalconnections.co.uk/download/help6/xdsobjects/XdsObjects.6.40~XdsObjects.Enums.XdsErrorCode.html
    private RegistryError createRegistryError(Exception e) {
        RegistryError error = new RegistryError();
        error.setErrorCode(XdsDocumentEntryConstants.XDS_ERROR_CODE_REPOSITORY_ERROR);
        if (e instanceof FSKException) {
            FSKException fskException = (FSKException) e;
            error.setCodeContext(fskException.getFaultCode() + ": " + fskException.getFaultInfo());
        } else {
            error.setCodeContext(e.getMessage());
        }
        return error;
    }
}
