/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.sds.dgws_idws_proxy.security.ProxyData;
import dk.sds.dgws_idws_proxy.security.SOSIUtil;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sdsd.nsp.slalog.ws.SLALoggingServletFilter;
import dk.stamkort.util.FaultCode;
import dk.stamkort.util.FaultUtil;
import dk.stamkort.util.RequestContextVO;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;
import java.io.StringWriter;
import java.util.Set;
import java.util.UUID;

public class DocumentProviderWSHandler implements SOAPHandler<SOAPMessageContext> {
    private static final Logger log = LogManager.getLogger(DocumentProviderWSHandler.class);

    private static SOSIUtil sosiUtil;

    @Override
    public Set<QName> getHeaders() {
        return null;
    }

    @Override
    public boolean handleMessage(SOAPMessageContext context) {
        Boolean outbound = (Boolean) context.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);

        if (outbound != null && !outbound) {
            SOAPMessage soapMessage = context.getMessage();

            Element securityHeader = null;
            Element medcomHeader = null;
            try {
                SOAPHeader soapHeader = soapMessage.getSOAPHeader();
                securityHeader = getHeader(soapHeader, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", "Security");
                medcomHeader = getHeader(soapHeader, "http://www.medcom.dk/dgws/2006/04/dgws-1.0.xsd", "Header");
            } catch (SOAPException e) {
                log.warn("DGWS headers not found in incoming message", e);
            }

            RequestContextVO requestContext = RequestContextVO.get();
            requestContext.setSecurityHeader(securityHeader);
            requestContext.setMedcomHeader(medcomHeader);
            requestContext.setProxyData(null);

            if (medcomHeader != null && securityHeader != null) {
                try {
                    String xml = getXML(soapMessage);
                    log.debug("Incoming SOAPMessage: " + xml);

                    ProxyData proxyData = sosiUtil.validateDGWS(xml, 3);
                    log.debug("ProxyData=" + proxyData);

                    // generate flowId if not set
                    String flowId = proxyData.getFlowId();
                    if (flowId == null) {
                        flowId = UUID.randomUUID().toString();
                        proxyData.setFlowId(flowId);
                    }
                    requestContext.setProxyData(proxyData);

                    try {
                        HttpServletRequest httpServletRequest = (HttpServletRequest)context.get(MessageContext.SERVLET_REQUEST);
                        log.debug("httpServletRequest=" + httpServletRequest);
                        SLALogItem slaLogItem = (SLALogItem) httpServletRequest.getAttribute(SLALoggingServletFilter.CURRENT_LOGITEM_REQUEST_KEY);
                        log.debug("slaLogItem=" + slaLogItem);
                        if (slaLogItem != null) {
                            slaLogItem.setMessageId(proxyData.getMessageId());
                        }
                        requestContext.setSlaLogItem(slaLogItem);
                    } catch (Exception e) {
                        log.warn("Unable to create SLALogItem", e);
                    }
                } catch (Exception e) {
                    FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER, e);
                }
            } else {
                FaultUtil.throwFSKFault(FaultCode.AUTHORIZATION_ERROR_INVALID_HEADER);
            }
        }

        return true;
    }

    private String getXML(SOAPMessage soapMessage) throws TransformerException, SOAPException {
        StringWriter sw = new StringWriter();
        Transformer t = TransformerFactory.newInstance().newTransformer();
        t.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
        t.transform(soapMessage.getSOAPPart().getContent(), new StreamResult(sw));

        String xml = sw.toString();

        if (xml != null) {
            xml = xml.replace("http://www.w3.org/2003/05/soap-envelope", "http://schemas.xmlsoap.org/soap/envelope/"); // change SOAP 1.2 to SOAP 1.1
        }

        return xml;
    }

    @Override
    public boolean handleFault(SOAPMessageContext context) {
        return true;
    }

    @Override
    public void close(MessageContext context) {
    }

    private Element getHeader(SOAPHeader soapHeader, String ns, String name) {
        if (soapHeader != null) {
            NodeList nodeList = soapHeader.getElementsByTagNameNS(ns, name);
            if (nodeList != null && nodeList.getLength() > 0) {
                log.debug("Got '" + name + "' header");
                return (Element) nodeList.item(0);
            }
        }
        return null;
    }

    public static void setSosiUtil(SOSIUtil s) {
        sosiUtil = s;
    }
}
