/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.conf;

import dk.cprsubscriber.dao.PersonSimplifiedDAO;
import dk.cprsubscriber.dao.impl.PersonSimplifiedDAOImpl;
import org.flywaydb.core.Flyway;
import org.junit.ClassRule;
import org.junit.rules.TemporaryFolder;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.jdbc.DataSourceProperties;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.Profile;
import org.springframework.jdbc.datasource.DataSourceTransactionManager;
import org.springframework.transaction.PlatformTransactionManager;

import javax.sql.DataSource;
import java.io.File;
import java.io.IOException;

@Configuration
@Profile("development")
public class TestDAOConfig {

    @ClassRule
    private TemporaryFolder tempDBFolder = new TemporaryFolder();
    TestDBScriptPreprocessor scriptPreprocessor = new TestDBScriptPreprocessor();

    private final Flyway flywayFSK = new Flyway();
    private final Flyway flywayStamdata = new Flyway();

    @Bean
    @Primary
    @ConfigurationProperties("spring.datasource")
    public DataSourceProperties dataSourceProperties() {
        return new DataSourceProperties();
    }

    @Bean
    @ConfigurationProperties("datasource-fsk")
    public DataSourceProperties fskDataSourceProperties() {
        return new DataSourceProperties();
    }

    @Bean
    @ConfigurationProperties("datasource-stm")
    public DataSourceProperties stamdataDataSourceProperties() {
        return new DataSourceProperties();
    }

    @Bean
    @Primary
    public DataSource fskDataSource(@Qualifier("fskDataSourceProperties") DataSourceProperties dataSourceProperties) {
        DataSource datasource = dataSourceProperties.initializeDataSourceBuilder().build();
        File scriptsLocation = prepareScripts("../compose/database/db/migration/fsk", "../compose/database/db/test/fsk");

        flywayFSK.setDataSource(datasource);
        flywayFSK.setLocations("filesystem:/" + scriptsLocation.getPath());
        flywayFSK.migrate();
        return datasource;
    }

    @Bean
    @Primary // make primary as we do not need transactions for other datasources
    public PlatformTransactionManager fskDataSourceTxManager(DataSource fskDataSource) {
        return new DataSourceTransactionManager(fskDataSource);
    }

    @Bean
    public DataSource stamdataDataSource(@Qualifier("stamdataDataSourceProperties") DataSourceProperties dataSourceProperties) {
        DataSource datasource = dataSourceProperties.initializeDataSourceBuilder().build();
        File scriptsLocation = prepareScripts("../compose/database/db/migration/stm", "../compose/database/db/test/stm");

        flywayStamdata.setDataSource(datasource);
        flywayStamdata.setLocations("filesystem:/" + scriptsLocation.getPath());
        flywayStamdata.migrate();
        return datasource;
    }

    @Bean
    public PlatformTransactionManager stamdataDataSourceTxManager(@Qualifier("stamdataDataSource") DataSource stamdataDataSource) {
        return new DataSourceTransactionManager(stamdataDataSource);
    }

    @Bean
    public PersonSimplifiedDAO personSimplifiedDAO(@Qualifier("stamdataDataSource") DataSource stamdataDataSource) {
        return new PersonSimplifiedDAOImpl(stamdataDataSource);
    }

    public Flyway getFlywayFSK() {
        return flywayFSK;
    }

    public Flyway getFlywayStamdata() {
        return flywayStamdata;
    }


    private File prepareScripts(String... paths) {
        try {
            tempDBFolder.create();
            File subFolder = tempDBFolder.newFolder();
            for (String path : paths) {
                scriptPreprocessor.preprocessScripts(path, subFolder);
            }
            return subFolder;
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }
}
