/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.integrations.dds;

import dk.cprsubscriber.dao.impl.CPRSubscriberTest;
import dk.stamkort.PersonalDataCardApplication;
import dk.stamkort.dao.PropertiesDAO;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.dao.exceptions.DDSException;
import dk.stamkort.dao.exceptions.DDSPermanentException;
import dk.stamkort.dao.exceptions.DDSTemporaryException;
import dk.stamkort.domain.PropertiesVO;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.util.Properties;
import dk.stamkort.util.SecretProperties;
import oasis.names.tc.ebxml_regrep.xsd.rs._3.RegistryResponseType;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.jdbc.core.namedparam.EmptySqlParameterSource;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import javax.sql.DataSource;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.HashMap;

import static dk.stamkort.integrations.dds.SyncJob.LAST_CPR_SYNC_TIME;
import static dk.stamkort.integrations.dds.SyncJob.PROPERTIES_DATE_FORMAT;
import static org.junit.Assert.*;
import static org.mockito.AdditionalAnswers.returnsFirstArg;
import static org.mockito.Matchers.anyString;
import static org.mockito.Mockito.*;

@RunWith(SpringRunner.class)
@SpringBootTest
@Transactional
public class SyncJobTest extends CPRSubscriberTest {

    private static final String CPR1 = "1111111111";
    private static final String CPR2 = "2222222222";
    private static final String CPR3 = "3333333333";
    private static final String CPR4 = "4444444444";

    protected NamedParameterJdbcTemplate jdbc;

    @Configuration
    static class TestConf extends PersonalDataCardApplication {
        @Bean public DDSRegisterOnDemandClient ddsOnDemandClient(Properties properties, SecretProperties secretProperties) { return mock(DDSRegisterOnDemandClient.class); }
        @Bean public DDSRegistryUpdateClient ddsRegistryUpdateClient(Properties properties, SecretProperties secretProperties) { return mock(DDSRegistryUpdateClient.class); }
        @Bean public DDSRegistryStoredQueryClient ddsRegistryStoredQueryClient(Properties properties, SecretProperties secretProperties) { return mock(DDSRegistryStoredQueryClient.class); }
    }

    @Autowired
    SyncJob syncJob;

    @Autowired
    private DataSource dataSource;

    @Autowired
    @Qualifier("stamdataDataSource")
    private DataSource stamdataDataSource;

    @Autowired
    private PropertiesDAO propertiesDAO;

    @Autowired
    private Properties properties;

    @Autowired
    RegistryIndexDAO registryIndexDAO;

    private NamedParameterJdbcTemplate stamdataJDBC() {
        return super.namedParamJdbcTemplate;
    }

    @Before
    public void setup() throws DAOException, DDSException {
        this.jdbc = new NamedParameterJdbcTemplate(dataSource);
        super.namedParamJdbcTemplate = new NamedParameterJdbcTemplate(stamdataDataSource);
        
        String propDate = PROPERTIES_DATE_FORMAT.format(ZonedDateTime.now().minusSeconds(6));
        propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDate);
        
        when(syncJob.ddsRegister.registerOnDemandDocument(anyString(), anyString())).thenReturn(new RegistryResponseType());
        when(syncJob.ddsQuery.getRegistryIdByFSKDocumentId(anyString())).then(returnsFirstArg()); // return passed parameter
        doNothing().when(syncJob.ddsUpdate).deprecateDocumentUsingDdsId(anyString(), anyString());
        String cprList = "('" + String.join("','", CPR1, CPR2, CPR3, CPR4) + "')";
        stamdataJDBC().update("DELETE FROM v2_Person where CPR in " + cprList, new EmptySqlParameterSource());
        jdbc.update("DELETE FROM RegistryIndex where PersonIdentifier in " + cprList, new EmptySqlParameterSource());
    }

    @After
    public void after() {
        // Resetting invocation count
        reset(syncJob.ddsQuery);
        reset(syncJob.ddsRegister);
        reset(syncJob.ddsUpdate);
    }

    private void createAlivePeople() {
        createPerson(CPR1, "01", toInstant("1935-01-01"), Instant.now(), toInstant("2999-12-31"));
        createPerson(CPR2, "01", toInstant("1975-05-05"), Instant.now(), toInstant("2999-12-31"));
        createPerson(CPR3, "01", toInstant("2001-12-24"), Instant.now(), toInstant("2999-12-31"));
    }

    private void registerDeath() {
        closeCprRecord(CPR1, Instant.now());
        createPerson(CPR1, "90", toInstant("1935-01-01"), Instant.now(), toInstant("2999-12-31"));
    }

    @Test
    public void testRepeatedJobRuns() throws Exception {

        // Check that repeated job executions cause expected table sizes
        logTableSizes();
        long indexStartCount = currentIndexCount();
        createAlivePeople();
        runSynchronizationJob();
        long actual = currentIndexCount();
        assertEquals(3 + indexStartCount, actual);
        assertEquals(0, indexMarkedForDeletionCount());

        verify(syncJob.ddsRegister, times(3)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());

        registerDeath();
        runSynchronizationJob();
        assertEquals(2 + indexStartCount, currentIndexCount());
        assertEquals(1, indexMarkedForDeletionCount());

        // Simulate delayed replication where modification date is before job ran the last time
        createPerson(CPR4, "01", toInstant("2018-01-01"), toInstant("2018-01-01"), toInstant("2999-12-31"));
        runSynchronizationJob();
        assertEquals(3 + indexStartCount, currentIndexCount());
        assertEquals(1, indexMarkedForDeletionCount());
    }

    @Test
    public void testDeathMoreThanOneYearAgoProvidesNoMetadata() throws InterruptedException, DDSException, DAOException {
    	
      String propDate = PROPERTIES_DATE_FORMAT.format(ZonedDateTime.now().minusYears(2));
      propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDate);

        // Check that a person who is deceased more than one year ago will not cause DDS metadata creation
        cleanFskTables();
        createPerson(CPR1, "01", toInstant("1975-01-01"), toInstant("1975-01-01"), toInstant("1976-01-01"));
        createPerson(CPR1, "01", toInstant("1976-01-01"), toInstant("1976-01-01"), toInstant("1977-01-01"));
        createPerson(CPR1, "01", toInstant("1977-01-01"), toInstant("1977-01-01"), toInstant("1978-01-01"));
        createPerson(CPR1, "01", toInstant("1978-01-01"), toInstant("1978-01-01"), toInstant("1979-01-01"));
        createPerson(CPR1, "90", toInstant("1979-01-01"), toInstant("1979-01-01"), toInstant("2999-12-31"));

        runSynchronizationJob();

        // Because the latest record has status 90, no DDS sync should be made
        verify(syncJob.ddsRegister, times(0)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }

    @Test
    public void testDeathLessThanOneYearAgoProvidesMetadataRegistration() throws InterruptedException, DDSException, DAOException {
    	
    	ZonedDateTime zdt = ZonedDateTime.now().minusSeconds(6); 
        String propDate = PROPERTIES_DATE_FORMAT.format(ZonedDateTime.now().minusYears(2));
        propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDate);

        // Check that a person who is deceased less than one year ago will cause DDS metadata creation
        cleanFskTables();
        Instant sixMonthsAgo = monthsAgo(6);
        createPerson(CPR1, "01", monthsAgo(24), monthsAgo(24), monthsAgo(18));
        createPerson(CPR1, "01", monthsAgo(18), monthsAgo(18), monthsAgo(12));
        createPerson(CPR1, "01", monthsAgo(12), monthsAgo(12), monthsAgo(6));
        createPerson(CPR1, "90", monthsAgo(6), monthsAgo(6), toInstant("2999-12-31"));

        runSynchronizationJob();

        // Because the latest record is only 6 months old, DDS sync should be called
        verify(syncJob.ddsRegister, times(1)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }

    @Test
    public void testMultipleRecordsForSamePerson() throws InterruptedException, DDSException, DAOException {
    	
      String propDate = PROPERTIES_DATE_FORMAT.format(ZonedDateTime.now().minusYears(2));
      propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDate);

        // Check that multiple records for the same person only cause one metadata registration. This may be the case
        // for eg the first run, where potentially many records may exist for the same CPR
        createPerson(CPR1, "01", monthsAgo(30), monthsAgo(30), monthsAgo(24));
        createPerson(CPR1, "01", monthsAgo(24), monthsAgo(24), monthsAgo(18));
        createPerson(CPR1, "01", monthsAgo(18), monthsAgo(18), monthsAgo(12));
        createPerson(CPR1, "01", monthsAgo(12), monthsAgo(12), toInstant("2999-12-31"));

        runSynchronizationJob();

        // Because the latest record is only 6 months old, DDS sync should be called
        verify(syncJob.ddsRegister, times(1)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }

    @Test
    public void testJobCleanup() throws DAOException, InterruptedException, DDSException {

        // Check that job cleanup cause expected deletions and callouts

        // Populate v2_Person and run the SyncJob to populate the index
        createAlivePeople();
        runSynchronizationJob();
        registerDeath();
        runSynchronizationJob();

        // Check that the "dead person record" is marked to be deleted
        RegistryIndexVO registryIndexVO = registryIndexDAO.getRegistryIndexVO(CPR1);
        assertNotNull(registryIndexVO);
        assertNotNull(registryIndexVO.getDeleteAfter());

        // Manipulate data such that DeleteAfter is more than one year ago
        String sql = "UPDATE RegistryIndex SET DeleteAfter = :deleteAfter WHERE PersonIdentifier = :cpr";
        final MapSqlParameterSource parameters = new MapSqlParameterSource();
        parameters.addValue("deleteAfter", "2017-01-01");
        parameters.addValue("cpr", CPR1);
        jdbc.update(sql, parameters);

        // Run job, which should now delete the index entry
        runSynchronizationJob();
        registryIndexVO = registryIndexDAO.getRegistryIndexVO(CPR1);
        assertNull(registryIndexVO);

        verify(syncJob.ddsRegister, times(3)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(1)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(1)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }

    @Test
    public void testJobErrorHandlingForTemporaryError() throws InterruptedException, DDSTemporaryException, DDSPermanentException {

        // The job should exit after the first failed attempt even though we configure it to accept 3 permanent failures.
        int maxNumberOfAcceptedErrors = 3;
        properties.setJobsDDSyncMaxErrors(maxNumberOfAcceptedErrors);

        // Simulate temporary DDS exception
        when(syncJob.ddsRegister.registerOnDemandDocument(anyString(), anyString())).thenThrow(new DDSTemporaryException(""));

        long indexStartCount = currentIndexCount();
        createAlivePeople();
        runSynchronizationJob();
        long actual = currentIndexCount();
        assertEquals(0 + indexStartCount, actual); // No new entries may be made
        assertEquals(0, indexMarkedForDeletionCount());

        // The job must exit after the first failed attempt
        verify(syncJob.ddsRegister, times(1)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }

    @Test
    public void testJobErrorHandlingForPermanentError() throws InterruptedException, DDSTemporaryException, DDSPermanentException {

        int maxNumberOfAcceptedErrors = 1;
        properties.setJobsDDSyncMaxErrors(maxNumberOfAcceptedErrors);

        // Simulate temporary DDS exception
        when(syncJob.ddsRegister.registerOnDemandDocument(anyString(), anyString())).thenThrow(new DDSPermanentException(""));

        long indexStartCount = currentIndexCount();
        createAlivePeople();
        runSynchronizationJob();
        long actual = currentIndexCount();
        assertEquals(0 + indexStartCount, actual); // No new entries may be made
        assertEquals(0, indexMarkedForDeletionCount());

        // The job must exit when the number of failed attempts exceed "maxNumberOfAcceptedErrors"
        verify(syncJob.ddsRegister, times(maxNumberOfAcceptedErrors + 1)).registerOnDemandDocument(anyObject(), anyObject());
        verify(syncJob.ddsQuery, times(0)).getRegistryIdByFSKDocumentId(anyObject());
        verify(syncJob.ddsUpdate, times(0)).deprecateDocumentUsingDdsId(anyObject(), anyObject());
    }
    
    @Test
    public void testGetAndSetOfDateProperties() throws DAOException  {
    	
    	ZonedDateTime zdtUpdate = ZonedDateTime.now();
    	String propDateUpdate = PROPERTIES_DATE_FORMAT.format(zdtUpdate);
    	
    	propertiesDAO.updateProperty(LAST_CPR_SYNC_TIME, propDateUpdate);
    	
    	PropertiesVO lastDeletionTime = propertiesDAO.getProperty(LAST_CPR_SYNC_TIME);
    	String propDateGet = lastDeletionTime.getPropertyValue();
    	
    	assertNotNull(lastDeletionTime);
    	assertEquals(propDateUpdate, propDateGet);
    	    	
    	LocalDateTime ldtGet = LocalDateTime.parse(propDateGet, PROPERTIES_DATE_FORMAT);
    	ZonedDateTime zdtGet = ldtGet.atZone(ZoneId.systemDefault());

    	assertEquals(zdtUpdate.withNano(0), zdtGet);

    }

    // UTILITY METHODS

    private Instant toInstant(String s) {
        LocalDate localDate = LocalDate.parse(s, DateTimeFormatter.ofPattern("yyyy-MM-dd"));
        return localDate.atStartOfDay().atZone(ZoneId.systemDefault()).toInstant();
    }

    private Instant monthsAgo(int months) {
        // Estimates a month to 30 days
        return Instant.now().minus(30 * months, ChronoUnit.DAYS);
    }

    private void runSynchronizationJob() throws InterruptedException {
        Thread.sleep(1000);
        syncJob.runSyncPersonIds();
    }

    private void logTableSizes() {
        long personCountBefore = personCount();
        long indexCountBefore = currentIndexCount();
        System.out.println("PersonCount="+ personCountBefore);
        System.out.println("IndexCount: "+ indexCountBefore);
    }

    private void cleanFskTables() {
        stamdataJDBC().update("DELETE FROM v2_Person", new EmptySqlParameterSource());
        jdbc.update("DELETE FROM RegistryIndex", new EmptySqlParameterSource());
    }

    private long personCount() {
        return stamdataJDBC().queryForObject("select count(1) from v2_Person where validFrom <= now() AND now() < validTo", new HashMap<>(), Long.class);
    }

    private long currentIndexCount() {
        return jdbc.queryForObject("select count(1) from RegistryIndex WHERE DeleteAfter IS NULL", new HashMap<>(), Long.class);
    }

    private long indexMarkedForDeletionCount() {
        return jdbc.queryForObject("select count(1) from RegistryIndex WHERE DeleteAfter IS NOT NULL", new HashMap<>(), Long.class);
    }

}
