/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.hl7cda;

import dk.oio.rep.cpr_dk.xml.schemas.core._2005._11._24.PersonBirthDateStructureType;
import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.CountryIdentificationCodeType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.RegisterData;
import dk.stamkort.webservice.hl7cda.model.CoverageGroupHolder;
import dk.stamkort.webservice.hl7cda.model.PDC_2_0DataHolder;
import dk.sundhedsdatastyrelsen.fsk._2019._08._01.GetPersonalDataCardResponse;
import oio.medcom.cprservice._1_0.*;
import org.junit.Test;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import java.io.InputStream;
import java.time.LocalDate;
import java.util.GregorianCalendar;

import static org.apache.logging.log4j.core.util.Assert.isNonEmpty;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

public class PDCDataMapperTest {

    @Test
    public void testSKREntryEmptyNoStructuredBody() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response_empty_structured_body.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setPersonalDataCard(personalDataCardResponse);
        registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().getCountryIdentificationCode().setValue("DK");

        YderVO yder = new YderVO();
        yder.setYdernr("123123");

        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        isNonEmpty(dataHolder);
        assertEquals(dataHolder.getPrimaryPersonInformation().getCpr(), "2512664949");
        assertEquals(dataHolder.getPrimaryPersonInformation().getAddress().getCountry(), "DK");
    }


    @Test
    public void testSKREntrySecretAdress() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setPersonalDataCard(personalDataCardResponse);
        registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().setPersonInformationProtectionIndicator(true);
        registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().setCountryIdentificationCode(null);

        YderVO yder = new YderVO();
        yder.setYdernr("123123");

        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        isNonEmpty(dataHolder);
        assertEquals(dataHolder.getPrimaryPersonInformation().getCpr(), "2512484916");
        assertEquals(dataHolder.getPrimaryPersonInformation().getAddress().getUse(), "CONF");
        assertNull(dataHolder.getPrimaryPersonInformation().getAddress().getCountry());
    }


    @Test
    public void testSKREntryHomeAdress() throws JAXBException, DatatypeConfigurationException {
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.setPersonalDataCard(personalDataCardResponse);
        registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getRegularCPRPerson().setPersonInformationProtectionIndicator(false);
        registerData.getScesData().getPersonWithHealthCareInformationStructure().getPersonInformationStructure().getPersonAddressStructure().getAddressComplete().getAddressPostal().setCountryIdentificationCode(null);

        YderVO yder = new YderVO();
        yder.setYdernr("123123");

        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        isNonEmpty(dataHolder);
        assertEquals(dataHolder.getPrimaryPersonInformation().getCpr(), "2512484916");
        assertEquals(dataHolder.getPrimaryPersonInformation().getAddress().getUse(), "H");
        assertNull(dataHolder.getPrimaryPersonInformation().getAddress().getCountry());

        // Relative phone
        assertEquals(3, dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().size());
        assertEquals("H", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(0).getUses().get(0));
        assertEquals("tel:98125342", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(0).getValue());
        assertEquals("MC", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(1).getUses().get(0));
        assertEquals("tel:25782345", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(1).getValue());
        assertEquals("WP", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(2).getUses().get(0));
        assertEquals("tel:null", dataHolder.getRelativeTypedInHolder().get(0).getPhoneHolders().get(2).getValue());
    }

    @Test
    public void testMapBirthTime() throws JAXBException, DatatypeConfigurationException {
        LocalDate birth = LocalDate.now();
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.getScesData()
                .getPersonWithHealthCareInformationStructure()
                .getPersonInformationStructure()
                .getRegularCPRPerson()
                .getPersonBirthDateStructure()
                .setBirthDate(DatatypeFactory.newInstance().newXMLGregorianCalendar(birth.toString()));

        YderVO yder = new YderVO();
        yder.setYdernr("123123");

        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        isNonEmpty(dataHolder);
        assertEquals(birth, dataHolder.getPrimaryPersonInformation().getBirthDate());
    }

    @Test
    public void testMapCoverageGroupHolder() throws JAXBException, DatatypeConfigurationException {
        CoverageGroupHolder expectedGroup = new CoverageGroupHolder();
        expectedGroup.setGroup("2");
        InputStream stream = getClass().getResourceAsStream("/testfiles/GetPersonalDataCard_response.xml");
        JAXBContext jaxbContext = JAXBContext.newInstance(GetPersonalDataCardResponse.class);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        GetPersonalDataCardResponse personalDataCardResponse = (GetPersonalDataCardResponse) unmarshaller.unmarshal(stream);

        String cprnr = personalDataCardResponse.getClinicalDocument().getRecordTargets().get(0).getPatientRole().getIds().get(0).getExtension(); // phew...

        RegisterData registerData = createRegisterData(cprnr);
        registerData.getScesData()
                .getPersonWithHealthCareInformationStructure()
                .getPersonHealthCareInformationStructure()
                .getPersonPublicHealthInsurance().setPublicHealthInsuranceGroupIdentifier(PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_2);

        YderVO yder = new YderVO();
        yder.setYdernr("123123");

        PDC_2_0DataHolder dataHolder = PDCDataMapper.mapDataHolder(registerData, yder);

        isNonEmpty(dataHolder);
        assertEquals(expectedGroup.getGroup(), dataHolder.getCoverageGroupHolder().getGroup());
    }

    private RegisterData createRegisterData(String cpr)  throws  DatatypeConfigurationException{
        RegisterData registerData = new RegisterData(cpr);

        RegularCPRPersonType regularCPRPersonType = new RegularCPRPersonType();
        regularCPRPersonType.setSimpleCPRPerson(new SimpleCPRPersonType());
        regularCPRPersonType.getSimpleCPRPerson().setPersonNameStructure(new PersonNameStructureType());

        regularCPRPersonType.setPersonBirthDateStructure(new PersonBirthDateStructureType());
        XMLGregorianCalendar xmlCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(new GregorianCalendar());
        xmlCal.setYear(2019);
        xmlCal.setMonth(1);
        xmlCal.setDay(1);
        regularCPRPersonType.getPersonBirthDateStructure().setBirthDate(xmlCal);

        PersonAddressStructureType personAddressStructureType = new PersonAddressStructureType();
        personAddressStructureType.setAddressComplete(new AddressCompleteType());
        personAddressStructureType.getAddressComplete().setAddressPostal(new AddressPostalType());
        personAddressStructureType.getAddressComplete().getAddressPostal().setCountryIdentificationCode(new CountryIdentificationCodeType());

        GetPersonWithHealthCareInformationOut scesData = new GetPersonWithHealthCareInformationOut();
        scesData.setPersonWithHealthCareInformationStructure(new PersonWithHealthCareInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().setPersonInformationStructure(new PersonInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().setRegularCPRPerson(regularCPRPersonType);
        scesData.getPersonWithHealthCareInformationStructure().getPersonInformationStructure().setPersonAddressStructure(personAddressStructureType);
        scesData.getPersonWithHealthCareInformationStructure().setPersonHealthCareInformationStructure(new PersonHealthCareInformationStructureType());
        scesData.getPersonWithHealthCareInformationStructure().getPersonHealthCareInformationStructure().setPersonPublicHealthInsurance(new PersonPublicHealthInsuranceType());

        registerData.setScesData(scesData);
        registerData.setHasOrganDonorRegistration(new Boolean(true));
        registerData.setHasTreatmentWill(new Boolean(true));
        registerData.setHasLivingWill(new Boolean(true));

        return registerData;
    }

}