/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2017 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.stamkort.webservice.ondemand;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.stamkort.dao.RegistryIndexDAO;
import dk.stamkort.dao.exceptions.DAOException;
import dk.stamkort.domain.RegistryIndexVO;
import dk.stamkort.domain.YderVO;
import dk.stamkort.integrations.*;
import dk.stamkort.service.PersonalDataCardServiceImpl;
import dk.stamkort.util.Properties;
import dk.stamkort.util.RequestContextVO;
import dk.stamkort.util.SecretProperties;
import dk.sundhedsdatastyrelsen.behandlingstestamente._2018._05._01.HasTreatmentWillResponse;
import dk.sundhedsdatastyrelsen.fsk._2019._08._01.GetPersonalDataCardResponse;
import dk.sundhedsdatastyrelsen.fsk.extensions.HealthProvider;
import dk.sundhedsdatastyrelsen.livstestamente._2018._05._01.HasLivingWillResponse;
import dk.sundhedsdatastyrelsen.organdonor._2018._05._01.HasOrganDonorRegistrationResponse;
import ihe.iti.xds_b._2007.RetrieveDocumentSetRequestType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType;
import ihe.iti.xds_b._2007.RetrieveDocumentSetResponseType.DocumentResponse;
import oio.medcom.cprservice._1_0.GetPersonWithHealthCareInformationOut;
import org.hl7.v3.*;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mockito;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.context.TestConfiguration;
import org.springframework.context.annotation.Bean;
import org.springframework.test.context.junit4.SpringRunner;
import org.springframework.transaction.annotation.Transactional;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.soap.SOAPException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.ws.Holder;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.List;
import java.util.UUID;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.mockito.Matchers.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@RunWith(SpringRunner.class)
@SpringBootTest
@Transactional
public class DocumentProviderWSTest {

    @TestConfiguration
    static class TestConf {
        @Bean
        public DocumentProviderWS documentProviderWS() {
            return new DocumentProviderWS();
        }

        @Bean
        public OrganDonorClient organDonorClient(Properties properties) {
            return mock(OrganDonorClient.class);
        }

        @Bean
        public LivingWillClient livingWillClient(Properties properties) {
            return mock(LivingWillClient.class);
        }

        @Bean
        public TreatmentWillClient treatmentWillClient(Properties properties) {
            return mock(TreatmentWillClient.class);
        }

        @Bean
        public PersonalDataCardRegisterClient personalDataCardRegisterClient(Properties properties) {
            return mock(PersonalDataCardRegisterClient.class);
        }

        @Bean
        public SCESClient scesClient(Properties properties, SecretProperties secretProperties) {
            return mock(SCESClient.class);
        }

        @Bean
        public YderClient yderClient(Properties properties) {
            return mock(YderClient.class);
        }
    }

    @Autowired
    RegistryIndexDAO indexDAO;

    @Autowired
    Properties properties;

    @Autowired
    SecretProperties secretProperties;

    @Autowired
    DocumentProviderWS ws;

    @Autowired
    public PersonalDataCardServiceImpl personalDataCardService;

    @Before
    public void setup() throws SOAPException, JAXBException, IOException, DGWSClientException, TransformerException {
        HasLivingWillResponse lwResponse = new HasLivingWillResponse();
        lwResponse.setWillExists(true);
        HasTreatmentWillResponse twResponse = new HasTreatmentWillResponse();
        twResponse.setWillExists(true);
        HasOrganDonorRegistrationResponse odResponse = new HasOrganDonorRegistrationResponse();
        odResponse.setRegistrationExists(false);
        when(personalDataCardService.livingWillClient.call(anyString(), anyObject())).thenReturn(lwResponse);
        when(personalDataCardService.treatmentWillClient.call(anyString(), anyObject())).thenReturn(twResponse);
        when(personalDataCardService.organDonorClient.call(anyString(), anyObject())).thenReturn(odResponse);

        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("123456"), personalDataCardService.cpre, properties, secretProperties);
        RequestContextVO.get().setProxyData(DocumentProviderTestData.sampleProxyData());

        YderVO yderResponse = DocumentProviderTestData.createYderData();
        when(DocumentProviderWS.documentProvider.personalDataCardService.yderClient.call(anyString(), anyObject())).thenReturn(yderResponse);
    }

    public static void mockCPREResponse(String response, SCESClient cpre, Properties p, SecretProperties secretProperties) throws IOException, JAXBException, SOAPException, DGWSClientException {
        Mockito.when(cpre.callInternal(anyObject(), anyString(), anyString())).thenReturn(response);
        Mockito.when(cpre.call(anyString(), any(RequestContextVO.class))).thenCallRealMethod();
        Mockito.when(cpre.setTransportUtil(any(TransportUtil.class))).thenCallRealMethod();
        Mockito.when(cpre.getUnmarshaller(GetPersonWithHealthCareInformationOut.class)).thenCallRealMethod();
        cpre.setTransportUtil(new TransportUtil(p.getClientKeystorePath(), secretProperties.getClientKeystorePassword(), "", "", "", 1, 1, "", false));
    }

    @Test
    public void testDocumentLookupNoYdernrSucces() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse(null);
        when(personalDataCardService.personalDataCardRegisterClient.call(anyString(), anyObject())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("0"), personalDataCardService.cpre, properties, secretProperties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");

        POCDMT000040PatientRole patientRole = clinicalDocument.getRecordTargets().get(0).getPatientRole();
        assertEquals(cpr, patientRole.getIds().get(0).getExtension());

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(4, entries.size());
        POCDMT000040Entry dentistEntry = entries.get(0);
        POCDMT000040Entry organDonorEntry = entries.get(1);
        POCDMT000040Entry livingWillEntry = entries.get(2);
        POCDMT000040Entry treatmentWillEntry = entries.get(3);

        assertNotNull(dentistEntry);
        HealthProvider healthProvider = dentistEntry.getHealthProvider();
        assertNotNull(healthProvider);
        POCDMT000040Organization organization = healthProvider.getOrganization();
        assertNotNull(organization);
        assertEquals(0, organization.getIds().size()); // No ydernr provided
        assertEquals("OrgName From Register", organization.getNames().get(0).getContent().get(0));
        assertEquals("Vej 42", ((AdxpStreetAddressLine) getAddrElement(organization, 0)).getContent().get(0).toString());
        assertEquals("8000", ((AdxpPostalCode) getAddrElement(organization, 1)).getContent().get(0).toString());
        assertEquals("Aarhus C", ((AdxpCity) getAddrElement(organization, 2)).getContent().get(0).toString());
        assertEquals("DK", ((AdxpCountry) getAddrElement(organization, 3)).getContent().get(0).toString());
        assertEquals(1, organization.getTelecoms().size()); // No email provided in SKR
        assertEquals("tel:11223344", organization.getTelecoms().get(0).getValue());

        // Check that the empty CPRE AssociatedGeneralPractitionerStructure causes no header participant
        assertEquals(0, clinicalDocument.getParticipants().size());

        checkSikringsgruppe(patientRole);
        checkOrganDonorAndLivingWillAndTreatmentWill(organDonorEntry, livingWillEntry, treatmentWillEntry);
    }

    private void checkSikringsgruppe(POCDMT000040PatientRole patientRole) {
        assertEquals("sygesikringsgruppe_1", patientRole.getTypeId().getExtension());
        assertEquals("1.2.208.184.15.12", patientRole.getTypeId().getRoot());
    }

    @Test
    public void testDocumentLookup_PDC_2_0_WithYdernrSucces() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse("123456");
        when(DocumentProviderWS.documentProvider.personalDataCardService.personalDataCardRegisterClient.call(anyString(), anyObject())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("123456"), DocumentProviderWS.documentProvider.personalDataCardService.cpre, properties, secretProperties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "PDC", "urn:ad:dk:medcom:pdc-v2.0:full");

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(8, entries.size());
    }


    @Test
    public void testDocumentLookupWithYdernrSucces() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse("123456");
        when(DocumentProviderWS.documentProvider.personalDataCardService.personalDataCardRegisterClient.call(anyString(), anyObject())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("123456"), DocumentProviderWS.documentProvider.personalDataCardService.cpre, properties, secretProperties);

        POCDMT000040ClinicalDocument clinicalDocument = getClinicalDocument(fskDocumentId, cpr, "52460-3", "DK FSK Schema");

        POCDMT000040PatientRole patientRole = clinicalDocument.getRecordTargets().get(0).getPatientRole();
        assertEquals(cpr, patientRole.getIds().get(0).getExtension());

        POCDMT000040Component3 component = clinicalDocument.getComponent().getStructuredBody().getComponents().get(0);
        List<POCDMT000040Entry> entries = component.getSection().getEntries();
        assertEquals(4, entries.size());
        POCDMT000040Entry dentistEntry = entries.get(0);
        POCDMT000040Entry organDonorEntry = entries.get(1);
        POCDMT000040Entry livingWillEntry = entries.get(2);
        POCDMT000040Entry treatmentWillEntry = entries.get(3);

        // Check that healthprovider has been enriched with data from yderregister
        assertNotNull(dentistEntry);
        HealthProvider healthProvider = dentistEntry.getHealthProvider();
        assertNotNull(healthProvider);
        POCDMT000040Organization organization = healthProvider.getOrganization();
        assertNotNull(organization);
        assertEquals("123456", organization.getIds().get(0).getExtension());
        assertEquals("Testyder", organization.getNames().get(0).getContent().get(0));
        assertEquals("Ydervej 42", ((AdxpStreetAddressLine) getAddrElement(organization, 0)).getContent().get(0).toString());
        assertEquals("1234", ((AdxpPostalCode) getAddrElement(organization, 1)).getContent().get(0).toString());
        assertEquals("Yderby", ((AdxpCity) getAddrElement(organization, 2)).getContent().get(0).toString());
        assertEquals("tel:12345678", organization.getTelecoms().get(0).getValue());
        assertEquals("mailto:test@testyder.dk", organization.getTelecoms().get(1).getValue());

        // Check that CPRE AssociatedGeneralPractitionerStructure is mapped to header participant and enriched with data from yderregister
        assertEquals(1, clinicalDocument.getParticipants().size());
        POCDMT000040Participant1 participant = clinicalDocument.getParticipants().get(0);
        assertEquals("1.3.6.1.4.1.19376.1.5.3.1.2.4", participant.getTemplateIds().get(0).getRoot());
        POCDMT000040AssociatedEntity associatedEntity = clinicalDocument.getParticipants().get(0).getAssociatedEntity();
        assertEquals("CAREGIVER", associatedEntity.getClassCodes().get(0));
        POCDMT000040Organization scopingOrganization = associatedEntity.getScopingOrganization();
        assertEquals("123456", scopingOrganization.getIds().get(0).getExtension());
        assertEquals("Testyder", scopingOrganization.getNames().get(0).getContent().get(0).toString());
        assertEquals("Testyder", scopingOrganization.getNames().get(0).getContent().get(0).toString());
        assertEquals("tel:12345678", scopingOrganization.getTelecoms().get(0).getValue());
        assertEquals("mailto:test@testyder.dk", scopingOrganization.getTelecoms().get(1).getValue());
        assertEquals("Ydervej 42", ((AdxpStreetAddressLine) getAddrElement(scopingOrganization, 0)).getContent().get(0).toString());
        assertEquals("1234", ((AdxpPostalCode) getAddrElement(scopingOrganization, 1)).getContent().get(0).toString());
        assertEquals("Yderby", ((AdxpCity) getAddrElement(scopingOrganization, 2)).getContent().get(0).toString());

        checkSikringsgruppe(patientRole);
        checkOrganDonorAndLivingWillAndTreatmentWill(organDonorEntry, livingWillEntry, treatmentWillEntry);
    }


    @Test
    public void testDocumentLookupWithStatus30() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("30");

        assertEquals("109: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }


    @Test
    public void testDocumentLookupWithStatus50() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("50");

        assertEquals("110: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }

    @Test
    public void testDocumentLookupWithStatus60() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("60");

        assertEquals("111: Det er ikke muligt at hente stamkortet, da CPR-nummeret er inaktivt: null", error);
    }

    @Test
    public void testDocumentLookupWithStatus90() throws DAOException, JAXBException, IOException, SOAPException, DGWSClientException {
        String error = testDocumentLookupWithStatusCommon("90");

        assertEquals("112: Det er ikke muligt at hente stamkortet, da borgeren er afgået ved døden for mere end 1 år siden: null", error);
    }

    private String testDocumentLookupWithStatusCommon(String status)
            throws DGWSClientException, IOException, JAXBException, SOAPException, DAOException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        // Setup PDC register mocked response with ydernr
        GetPersonalDataCardResponse skrResponse = DocumentProviderTestData.createPersonalDataCardResponse("123456");
        when(DocumentProviderWS.documentProvider.personalDataCardService.personalDataCardRegisterClient.call(anyString(), anyObject())).thenReturn(skrResponse);
        mockCPREResponse(DocumentProviderTestData.stubbedCPRE("123456", status), DocumentProviderWS.documentProvider.personalDataCardService.cpre, properties, secretProperties);

        return getClinicalDocumentError(fskDocumentId, cpr);
    }


    private void checkOrganDonorAndLivingWillAndTreatmentWill(POCDMT000040Entry organDonorEntry, POCDMT000040Entry livingWillEntry, POCDMT000040Entry treatmentWillEntry) {
        assertNotNull(organDonorEntry);
        assertEquals("OrganDonor", organDonorEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.15.1", organDonorEntry.getObservation().getCode().getCodeSystem());
        assertEquals("OrganDonorRegistration", organDonorEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, organDonorEntry.getObservation().getValues().size());
        assertEquals(Boolean.FALSE, ((BN) organDonorEntry.getObservation().getValues().get(0)).isValue());

        assertNotNull(livingWillEntry);
        assertEquals("LivingWill", livingWillEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.15.1", livingWillEntry.getObservation().getCode().getCodeSystem());
        assertEquals("Living Will", livingWillEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, livingWillEntry.getObservation().getValues().size());
        assertEquals(Boolean.TRUE, ((BN) livingWillEntry.getObservation().getValues().get(0)).isValue());

        assertNotNull(treatmentWillEntry);
        assertEquals("TreatmentWill", treatmentWillEntry.getObservation().getCode().getCode());
        assertEquals("1.2.208.184.15.1", treatmentWillEntry.getObservation().getCode().getCodeSystem());
        assertEquals("Treatment Will", treatmentWillEntry.getObservation().getCode().getDisplayName());
        assertEquals(1, treatmentWillEntry.getObservation().getValues().size());
        assertEquals(Boolean.TRUE, ((BN) treatmentWillEntry.getObservation().getValues().get(0)).isValue());
    }

    private Object getAddrElement(POCDMT000040Organization organization, int elementIndex) {
        JAXBElement actual = (JAXBElement) organization.getAddrs().get(0).getContent().get(elementIndex);
        return actual.getValue();
    }

    private POCDMT000040ClinicalDocument getClinicalDocument(String fskDocumentId, String cpr, String typeCode, String formatCode) throws DAOException, JAXBException {
        RetrieveDocumentSetResponseType response = getClinicalDocumentResponse(fskDocumentId, cpr, typeCode, formatCode);
        List<DocumentResponse> documentResponse = response.getDocumentResponse();
        RetrieveDocumentSetResponseType.DocumentResponse responseDocument = documentResponse.get(0);
        assertEquals("Returned documentId must match requested document id", fskDocumentId, responseDocument.getDocumentUniqueId());
        assertEquals(properties.getDdsHomeCommunityId(), responseDocument.getHomeCommunityId());
        assertNotNull(responseDocument.getDocument());
        System.out.println(new String(responseDocument.getDocument()));

        JAXBContext context = JAXBContext.newInstance(POCDMT000040ClinicalDocument.class);
        Unmarshaller unmarshaller = context.createUnmarshaller();
        ByteArrayInputStream bais = new ByteArrayInputStream(responseDocument.getDocument());
        // https://www.igorkromin.net/index.php/2017/03/15/unmarshalling-an-xml-fragment-representing-a-jaxb-object-without-xmlrootelement/
        return unmarshaller.unmarshal(new StreamSource(bais), POCDMT000040ClinicalDocument.class).getValue();
    }

    private String getClinicalDocumentError(String fskDocumentId, String cpr) throws DAOException, JAXBException {
        RetrieveDocumentSetResponseType response = getClinicalDocumentResponse(fskDocumentId, cpr, "52460-3", "DK FSK Schema");

        return response.getRegistryResponse().getRegistryErrorList().getRegistryError().get(0).getCodeContext();
    }

    private RetrieveDocumentSetResponseType getClinicalDocumentResponse(String fskDocumentId, String cpr, String typeCode, String formatCode)
            throws DAOException {
        RegistryIndexVO indexVO = new RegistryIndexVO();
        indexVO.setPersonIdentifier(cpr);
        indexVO.setFSKDocumentId(fskDocumentId);
        indexVO.setTypeCode(typeCode);
        indexVO.setFormatCode(formatCode);
        indexDAO.createRegistryIndexVO(indexVO);

        Security securityHeader = null;
        Holder<Header> medcomHeader = null;
        HsuidHeader hsuidHeader = null;
        RetrieveDocumentSetRequestType body = new RetrieveDocumentSetRequestType();
        RetrieveDocumentSetRequestType.DocumentRequest documentRequest = new RetrieveDocumentSetRequestType.DocumentRequest();
        documentRequest.setDocumentUniqueId(fskDocumentId);
        documentRequest.setHomeCommunityId(properties.getDdsHomeCommunityId());
        documentRequest.setRepositoryUniqueId(properties.getDdsRepositoryUniqueId());
        body.getDocumentRequest().add(documentRequest);
        return ws.documentRepositoryRetrieveDocumentSet(body, securityHeader, medcomHeader, hsuidHeader);
    }


    @Test
    public void testDocumentLookupForNonExistingDocumentid() throws DAOException {
        String fskDocumentId = UUID.randomUUID().toString();
        String cpr = "1122334455";

        RegistryIndexVO indexVO = new RegistryIndexVO();
        indexVO.setPersonIdentifier(cpr);
        indexVO.setFSKDocumentId(fskDocumentId);
        indexDAO.createRegistryIndexVO(indexVO);

        Security securityHeader = null;
        Holder<Header> medcomHeader = null;
        HsuidHeader hsuidHeader = null;
        RetrieveDocumentSetRequestType body = new RetrieveDocumentSetRequestType();
        RetrieveDocumentSetRequestType.DocumentRequest documentRequest = new RetrieveDocumentSetRequestType.DocumentRequest();
        documentRequest.setDocumentUniqueId("dummy-foo-bar");
        documentRequest.setHomeCommunityId(properties.getDdsHomeCommunityId());
        documentRequest.setRepositoryUniqueId(properties.getDdsRepositoryUniqueId());
        body.getDocumentRequest().add(documentRequest);
        RetrieveDocumentSetResponseType response = ws.documentRepositoryRetrieveDocumentSet(body, securityHeader, medcomHeader, hsuidHeader);
        System.out.println(response);
        assertEquals("Did not expect any returned document for dummy document id", 0, response.getDocumentResponse().size());
    }

}
